package EFS::CLI::Campus;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use Params::Validate qw( :all );

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region      => { type => SCALAR },
        name        => { type => SCALAR },
        description => { type => SCALAR },
    });

    my $region_rs = $class->get_resultset( q{Region} );

    my $region = $region_rs->retrieve_bynames_assert(
        name => $args{region},
    );

    my $resultset = $class->get_resultset;

    my ($campus) = $resultset->search({ name => $args{name} });

    if ( $campus ) {

        if ( $campus->region->name ne $args{region} ) {
            die(
                qq{Campus $args{name} is already defined, but in },
                q{region }, $campus->region->name, qq{\n},
            );
        } else {
            STDOUT->print(
                qq{Campus $args{region}/$args{name} is already defined\n},
            );
        }

        if ( $campus->description ne $args{description} ) {
            $campus->set_column( description => $args{description} );
            $campus->update;
            STDOUT->print( qq{Updated description of campus $campus\n} );
        }

    } else {

        $campus = $resultset->create({
            region      => $region->id,
            name        => $args{name},
            description => $args{description},
        });

        STDOUT->print( qq{Defined campus $args{region}/$args{name}\n} );

    }

    return $campus;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region => { type => SCALAR },
        name   => { type => SCALAR },
    });

    my $campus_rs = $class->get_resultset( q{Campus} );

    my $campus = $campus_rs->retrieve_bynames_assert(
        region => $args{region},
        name   => $args{name},
    );

    if ( my @locations  = $campus->locations ) {
        die(
            qq{Unable to undef campus $campus, because it has\n},
            qq{the following locations:\n\t},
            join(qq{\n\t}, map { $_->name } @locations ), qq{\n},
        );
    }

    my $object_id = $campus->object_id;

    $campus->attrs->delete_all;
    $campus->delete;

    STDOUT->print( qq{Undefined campus $object_id\n} );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region => { type     => SCALAR,
                    optional => 1 },
        name   => { type     => SCALAR,
                    depends  => [qw( region )],
                    optional => 1 },
    });

    my $resultset = $class->get_resultset;

    if ( not $args{region} ) {
        $class->display_objects( objects => [ $resultset->all ] );
        return 1;
    }

    my $region_rs = $class->get_resultset( q{Region} );

    my $region = $region_rs->retrieve_bynames_assert(
        name => $args{region},
    );

    if ( not $args{name} ) {
        if ( my @campuses = $region->campuses ) {
            $class->display_objects( objects => \@campuses );
        } else {
            STDOUT->print(
                qq{No campuses defined for region $args{region}\n},
            );
        }
        return 1;
    }

    my $campus = $resultset->retrieve_bynames_assert(
        region => $args{region},
        name   => $args{name},
    );

    $class->display_objects( objects => [ $campus ] );

    if ( my @attrs = $campus->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( my @locations = $campus->locations ) {
        $locations[0]->display_objects( objects => \@locations );
    }

    if ( my @cells = $campus->cells ) {
        $cells[0]->display_objects( objects => \@cells );
    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region => { type => SCALAR },
        name   => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{campus} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
