package EFS::CLI::Cell;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use Readonly;
use Params::Validate qw( :all );
use Try::Tiny;

use EFS::FQDN qw(fqdn);

Readonly my $SCREEN_WIDTH => 1024;

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region     => { type => SCALAR },
        campus     => { type => SCALAR },
        location   => { type => SCALAR },
        name       => { type => SCALAR },
        celltype   => { type => SCALAR },
        fstype     => { type => SCALAR },
        parentcell => { type => SCALAR, optional => 1 },
    });

    my $location_rs = $class->get_resultset( q{Location} );

    my $location = $location_rs->retrieve_bynames_assert(
        region => $args{region},
        campus => $args{campus},
        name   => $args{location},
    );

    my $location_name = $location->name;

    my $resultset = $class->get_resultset;

    $args{name} = $resultset->expand_cell_name( %args );

    my $parentcell;

    if ( $args{parentcell} ) {

        if ( $args{celltype} eq q{dev} ) {
            die(
                qq{Invalid argument combination.\n},
                qq{Dev cells can not have parentcells.\n}
            );
        }

        $args{parentcell} = $resultset->expand_cell_name( name => $args{parentcell} );

        $parentcell = $resultset->retrieve_bynames_assert( name => $args{parentcell} );

    } else {

        if ( $args{celltype} ne q{dev} ) {
            die(
                qq{Missing required argument 'parentcell'.\n},
                qq{A parentcell must be specified for test or prod cells.\n}
            );
        }

    }

    my ($cell) = $resultset->search({
        location => $location->id,
        name     => $args{name},
    });

    if ( $cell ) {

        if ( $cell->location->name ne $args{location} ||
                 $cell->celltype ne $args{celltype} ||
                     $cell->fstype ne $args{fstype} ) {
            die(
                qq{Cell $args{name} already exists with:\n},
                q{Location => }, $cell->location->name, qq{\n},
                q{Celltype => }, $cell->celltype, qq{\n},
                q{FSType   => }, $cell->fstype, qq{\n},
                qq{These can NOT be redefined.\n},
            );
        }

        if ( $cell->parentcell ) {
            if ( $cell->parentcell->name eq $args{parentcell} ) {
                STDOUT->print(
                    qq{Cell $args{name} is already defined with:\n},
                    qq{Parentcell => $args{parentcell}\n},
                );
            } else {
                $cell->set_column( parentcell => $parentcell->id );
                $cell->update;
                STDOUT->print(
                    qq{Updated parentcell for $args{name} to $args{parentcell}\n},
                );
            }
        } else {
            STDOUT->print( qq{Cell $args{name} already exists.\n} );
        }

    } else {

        my $create = {
            location => $location->id,
            name     => $args{name},
            celltype => $args{celltype},
            fstype   => $args{fstype},
            hidden   => 1,
        };

        if ( $parentcell ) {
            $create->{parentcell} = $parentcell->id;
        }

        $cell = $resultset->create( $create );

        STDOUT->print(
            qq{Defined cell '$args{name}' in location '$location_name', },
            qq{celltype '$args{celltype}'},
            ( $args{parentcell} ? qq{, parentcell '$args{parentcell}'.\n} : qq{\n} )
        );

    }

    return $cell;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $cell = $resultset->retrieve_bynames_assert( name => $args{name} );

    foreach my $childtype ( qw( fileservers efsservers projects ) ) {
        if ( my @childobjects = $cell->$childtype ) {
            die(
                qq{Unable to undef cell '$args{name}', },
                qq{as it has the following $childtype:\n\t},
                join(qq{\n\t}, map { $_->object_id } @childobjects ), qq{\n},
            );
        }
    }

    $cell->efsserver_cell->delete_all;
    $cell->attrs->delete_all;
    $cell->delete;

    STDOUT->print( qq{Undefined cell '$args{name}'\n} );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR, optional => 1 },
        # Use this one, since it defaults to celltype => all
        %EFS::CLI::validate_diststate_args,
    });

    my @display_columns = qw( region campus location name
                              celltype fstype parentcell disabled );

    if ( $args{hiddencells} ) {
        push @display_columns, q{hidden};
    }

    my $resultset = $class->get_resultset;

    if ( not $args{name} ) {

        my @allcells = $class->handle_cell_arguments(
            %args,
            default_scope => q{global},
        );

        foreach my $celltype ( qw( test dev prod ) ) {
            if ( my @cells = grep { $_->celltype eq $celltype } @allcells ) {
                STDOUT->print(qq{\n} . ucfirst($celltype) . qq{ cells:\n});
                $class->display_objects(
                    objects         => \@cells,
                    display_columns => \@display_columns,
                );
            }
        }

        return 1;

    }

    my $cell = $resultset->retrieve_bynames_assert( name => $args{name} );

    $class->display_objects(
        objects         => [ $cell ],
        display_columns => \@display_columns,
    );

    if ( my @attrs = $cell->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( my @fileservers = $cell->fileservers ) {
        $fileservers[0]->display_objects( objects => \@fileservers );
    }

    if ( my @efsservers = $cell->efsservers ) {
        $efsservers[0]->display_objects( objects => \@efsservers );
    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{cell} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

sub dumpconfig {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_dumpconfig_args,
    });

    my $project_rs = $class->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => q{efs},
        name     => q{config-automaps},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{dumpconfig_rpc},
        args   => [ %args ],
    );

}

sub dumpconfig_local {

    my $class = shift;

    my $config = EFS::Config::Server->new;

    my $type = $config->get_value( q{server}, q{automounter} ) || q{autofs};

    my $install_dir = qq{/var/efs/etc/automaps/$type};

    EFS::Util->MakeDirHierarchy(
        top    => q{/var/efs},
        bottom => $install_dir,
    );

    my $dirlock = EFS::DirLock->new({
        directory => $install_dir,
        retry     => 1,
        command   => q{dumpconfig_local},
    });

    my $efsserver_rs = $class->get_resultset( q{EFSServer} );

    my $efsserver = $efsserver_rs->retrieve_bynames_assert( name => fqdn );

    foreach my $cell ( $efsserver->cells ) {
        next if $cell->fstype eq q{afs};
        if ( $type eq q{autofs} ) {
            $cell->_generate_cell_auto_map_global( $install_dir );
            $cell->_generate_cell_auto_map_admin( $install_dir );
        } else {
            $cell->_generate_cell_amd_map_global( $install_dir );
            $cell->_generate_cell_amd_map_admin( $install_dir );
        }
    }

    STDOUT->print( qq{Updated locally installed $type maps\n} );

    return 1;

}

sub dumpconfig_rpc {

    my $class = shift;

    my %args = @_;

    my $history_rs = $class->get_resultset( q{History} );
    my $history    = $history_rs->retrieve_assert( id => $args{history} );

    my $release = EFS::CLI::AutoRelease->create_rpc(
        history  => $history->id,
        metaproj => q{efs},
        project  => q{config-automaps},
    );

    my ($install) = $release->installs({ name => q{common} });

    my $install_dir = $install->get_path(q{dev});

    my $resultset = $class->get_resultset;

    my @cells = $resultset->search({ celltype => q{dev}, fstype => q{nfs} });

    foreach my $cell ( @cells ) {
        $cell->_generate_cell_auto_map_global( qq{$install_dir/autofs} );
        $cell->_generate_cell_auto_map_admin( qq{$install_dir/autofs} );
    }

    foreach my $cell ( @cells ) {
        $cell->_generate_cell_amd_map_global( qq{$install_dir/amd} );
        $cell->_generate_cell_amd_map_admin( qq{$install_dir/amd} );
    }

    EFS::CLI::AutoRelease->dist_rpc(
        history  => $history->id,
        metaproj => q{efs},
        project  => q{config-automaps},
    );

    STDOUT->print( qq{Updated automounter maps\n} );

    return 1;

}

sub clonecellstate {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        fromcell => { type => SCALAR },
        tocell   => { type => SCALAR },
        force    => { type => SCALAR, optional => 1 },
    });

    $class->assert_maintenance_mode;

    if ( not $args{force} ) {
        $class->assert_no_active_dists;
    }

    my $resultset = $class->get_resultset;

    my $fromcell =
        $resultset->retrieve_bynames_assert( name => $args{fromcell} );
    my $tocell =
        $resultset->retrieve_bynames_assert( name => $args{tocell} );

    if ( $fromcell->name eq $tocell->name ) {
        die qq{Cell names may not match\n};
    }

    if ( $fromcell->celltype ne $tocell->celltype ) {
        die qq{Cell types must match\n};
    }

    if ( $fromcell->hidden ) {
        die qq{From cell must NOT be hidden\n};
    }

    if ( not $tocell->hidden ) {
        die qq{To cell must be hidden\n};
    }

    my $install_rs      = $class->get_resultset( q{InstallCellState} );
    my $releasealias_rs = $class->get_resultset( q{ReleaseAliasCellState} );

    my $schema = $class->get_schema;

    my $transaction = sub {

        my $install_to_rs = $install_rs->search({ cell => $tocell->id });

        if ( my $count = $install_to_rs->count ) {
            STDOUT->print( qq{Deleting $count install records for $tocell\n} );
            $install_to_rs->delete_all;
        }

        my $install_from_rs = $install_rs->search({ cell => $fromcell->id });

        if ( my $count = $install_from_rs->count ) {

            STDOUT->print( qq{Cloning $count install records for $tocell\n} );

            my @data = map { {
                cell       => $tocell->id,
                install    => $_->install->id,
                updatetime => $_->updatetime,
            } } $install_from_rs->all;

            my @results = $install_rs->populate( \@data );

        }

        my $releasealias_to_rs = $releasealias_rs->search({ cell => $tocell->id });

        if ( my $count = $releasealias_to_rs->count ) {
            STDOUT->print( qq{Deleting $count releasealias records for $tocell\n} );
            $releasealias_to_rs->delete_all;
        }

        my $releasealias_from_rs = $releasealias_rs->search({ cell => $fromcell->id });

        if ( my $count = $releasealias_from_rs->count ) {

            STDOUT->print( qq{Cloning $count releasealias records for $tocell\n} );

            my @data = map { {
                cell         => $tocell->id,
                releasealias => $_->releasealias->id,
                target       => $_->target,
            } } $releasealias_from_rs->all;

            my @results = $releasealias_rs->populate( \@data );

        }

    };

    my $eval_error;
    try {
        $schema->txn_do( $transaction );
    }
    catch {
        $eval_error = $_;
    };


    if ( $eval_error ) {
        die(
            q{Unable to clone cellstate },
            qq{from $fromcell to $tocell: $eval_error\n},
        );
    }

    STDOUT->print( qq{Cloned cell state from $fromcell to $tocell\n} );

    return 1;

}

# XXX: This needs to take ReleaseAliasCellStates into account!!!
sub diffcellstate {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_dist_args,
    });

    my @cells = $class->handle_cell_arguments(
        %args,
        default_scope => q{global},
    );

    my %cells_bytype = ();
    my %cells_byname = ();

    foreach my $cell ( @cells ) {
        $cells_bytype{ $cell->celltype }->{ $cell->name } = $cell;
        $cells_byname{ $cell->name } = $cell;
    }

    if ( scalar keys %cells_bytype > 1 ) {
        my $fatal = qq{All cells must have the same celltype\n};
        foreach my $celltype ( sort keys %cells_bytype ) {
            $fatal .= qq{\t$celltype\t=> } .
                join( q{,},
                      sort map { $_->shortname }
                          values %{ $cells_bytype{ $celltype } } ) . qq{\n};
        }
        die $fatal;
    }

    STDOUT->print( qq{Querying all install cellstates...\n} );

    my $cellstate_rs = $class->get_resultset( q{InstallCellState} );

    my @cellstates = $cellstate_rs->all;

    STDOUT->print( qq{Analyzing all install attributes...\n} );

    my %installs = ();

    foreach my $cellstate ( @cellstates ) {

        my $cell_name = $cellstate->cell->name;

        if ( not $cells_byname{ $cell_name } ) {
            next;
        }

        $installs{ $cellstate->install->object_id }->{ $cell_name } =
            $cellstate->updatetime;

    }

    STDOUT->print( qq{Generating output table...\n} );

    my $table = Text::FormatTable->new(
        q{l} . ( q{  r} x scalar keys %cells_byname )
    );

    $table->head( q{Install},
                  map { $cells_byname{$_}->shortname }
                      sort keys %cells_byname );
    $table->rule( q{=} );

    my $total = 0;
    my $diffs = 0;

    foreach my $id_mpri ( sort keys %installs ) {

        $total++;

        my %values    = ();
        my $max_value = 0;

        foreach my $cell_name ( sort keys %cells_byname ) {
            my $value = $installs{ $id_mpri }->{ $cell_name } || 0;
            $values{ $value }++;
            if ( $value > $max_value ) {
                $max_value = $value;
            }
        }

        if ( scalar keys %values == 1 ) {
            next;
        }

        my @row = ( $id_mpri );

        foreach my $cell_name ( sort keys %cells_byname ) {
            my $value = $installs{ $id_mpri }->{ $cell_name } || 0;
            if ( $value == $max_value ) {
                push @row, color_warn( $value );
            } else {
                push @row, $value;
            }
        }

        $table->row( @row );

        $diffs++;

    }

    if ( $diffs ) {
        STDOUT->print(
            qq{Found $diffs out of $total installs with different diststates:\n},
            $table->render($SCREEN_WIDTH),
        );
    } else {
        STDOUT->print(
            qq{Both cells are completely in sync\n},
        );
    }

    return 1;

}

sub getnames {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        # Use this one, since it defaults to celltype => all
        %EFS::CLI::validate_diststate_args,
    });

    my $resultset = $class->get_resultset;

    my @cells = $class->handle_cell_arguments( %args, default_scope => q{global} );

    STDOUT->print( map { $_->name . qq{\n} } @cells );

    return 1;

}

sub hide {

    my $class = shift;

    my %args = validate( @_, {
        %EFS::CLI::validate_common,
        name  => { type => SCALAR },
    });

    my $cell = $class->retrieve_bynames_assert( name => $args{name} );

    if ( my @children = grep { not $_->hidden } sort $cell->children ) {
        die(
            qq{The following child cells are not yet hidden:\n\t},
            join( qq{\n\t}, @children ), qq{\n},
            qq{Unable to hide cell $cell\n},
        );
    }

    $cell->set_column( hidden => 1 );
    $cell->update;

    STDOUT->print(qq{Cell $cell has been hidden\n});

    return 1;

}

sub expose {

    my $class = shift;

    my %args = validate( @_, {
        %EFS::CLI::validate_common,
        name  => { type => SCALAR },
    });

    my $cell = $class->retrieve_bynames_assert( name => $args{name} );

    if ( my $parentcell = $cell->parentcell ) {
        if ( $parentcell->hidden ) {
            die(
                qq{Parentcell $parentcell is hidden\n},
                qq{Unable to expose $cell\n},
            );
        }
    }

    $cell->set_column( hidden => 0 );
    $cell->update;

    STDOUT->print(qq{Cell $cell has been exposed\n});

    return 1;

}

sub execlinks {

    my $class = shift;

    my (%args)          = validate( @_, {
        %EFS::CLI::validate_common,
        name  => { type => SCALAR },
        force => { type => SCALAR | UNDEF, optional => 1 },
        %EFS::CLI::validate_diststate_args,
    });

    $class->assert_maintenance_mode;

    my $resultset = $class->get_resultset;
    my $cell = $resultset->retrieve_bynames_assert( name => $args{name} );

    $cell->assert_fstype( q{nfs} );

    return EFS::RPC->execute(
        cell   => $cell,
        class  => $class,
        method => q{execlinks_rpc},
        args   => [ %args ],
    );

}

sub execlinks_rpc {

    my $class = shift;
    my %args = @_;

    my $resultset = $class->get_resultset;
    my $cell = $resultset->retrieve_bynames_assert( name => $args{name} );

    my @fileqtrees = sort grep { $_->has_exectype } $cell->fileqtrees;

    my $fm_fileqtree = EFS::ForkManager->new({

        list => \@fileqtrees,

        action => sub {

            my $fileqtree = shift;

            return EFS::CLI::Fileqtree->execlinks_rpc(
                cell       => $fileqtree->cell->name,
                fileserver => $fileqtree->fileserver->name,
                fileshare  => $fileqtree->fileshare->name,
                name       => $fileqtree->name,
                force      => $args{force},
            );

        },

    });

    $fm_fileqtree->run || die(
        qq{Unable to update execlinks on the following fileqtrees:\n\t},
        join( qq{\n\t}, map { $_->object_id } $fm_fileqtree->get_failed ),
        qq{\n},
    );

    STDOUT->print( qq{Updated execlinks on all fileqtrees\n} );

    return 1;

}

sub init {

    my $class = shift;

    my (%args)          = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR },
        %EFS::CLI::validate_diststate_args,
    });

    my $resultset = $class->get_resultset;
    my $cell = $resultset->retrieve_bynames_assert( name => $args{name} );

    $cell->assert_fstype( q{afs} );

    return EFS::RPC->execute(
        cell   => $cell,
        class  => $class,
        method => q{init_rpc},
        args   => [ %args ],
    );

}

sub init_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;
    my $cell = $resultset->retrieve_bynames_assert( name => $args{name} );

    my @partitions = $cell->get_partition_list;

    EFS::AFS::Util->create_volume(
        name       => q{efs.root},
        cell       => $cell->name,
        partitions => \@partitions,
        readonly   => 1,
    );

    EFS::AFS::Util->create_mount(
        volume => q{efs.root},
        parent => q{root.cell},
        path   => q{efs},
        cell   => $cell->name,
    );

    my @types = qw( dist );

    if ( $cell->celltype ne q{prod} ) {
        push @types, qw( dev home global );
    }

    foreach my $type ( @types ) {
        EFS::AFS::Util->create_volume(
            name       => qq{efs.$type},
            cell       => $cell->name,
            partitions => \@partitions,
            readonly   => 1,
        );
        EFS::AFS::Util->create_mount(
            volume => qq{efs.$type},
            parent => q{efs.root},
            path   => $type,
            cell   => $cell->name,
        );
    }    

    if ( $cell->celltype ne q{prod} ) {
        foreach my $type ( qw( dev home ) ) {
            EFS::AFS::Util->create_volume(
                name       => qq{efs.global.$type},
                cell       => $cell->name,
                partitions => \@partitions,
                readonly   => 1,
            );
            EFS::AFS::Util->create_mount(
                volume => qq{efs.global.$type},
                parent => q{efs.global},
                path   => $type,
                cell   => $cell->name,
            );
        }    
    }

    return 1;

}

sub uninit {

    my $class = shift;

    my (%args)          = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR },
        %EFS::CLI::validate_diststate_args,
    });

    my $resultset = $class->get_resultset;
    my $cell = $resultset->retrieve_bynames_assert( name => $args{name} );

    $cell->assert_fstype( q{afs} );

    return EFS::RPC->execute(
        cell   => $cell,
        class  => $class,
        method => q{uninit_rpc},
        args   => [ %args ],
    );

}

sub uninit_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;
    my $cell = $resultset->retrieve_bynames_assert( name => $args{name} );

    if ( not $cell->hidden ) {
        die qq{Cell $cell is not hidden\n};
    }

    if ( $cell->celltype ne q{prod} ) {
        foreach my $type ( qw( dev home ) ) {
            EFS::AFS::Util->remove_mount(
                volume => qq{efs.global.$type},
                parent => q{efs.global},
                path   => $type,
                cell   => $cell->name,
            );
            EFS::AFS::Util->remove_volume(
                name => qq{efs.global.$type},
                cell => $cell->name,
            );
        }    
    }

    my @types = qw( dist );

    if ( $cell->celltype ne q{prod} ) {
        push @types, qw( dev home global );
    }

    foreach my $type ( @types ) {
        EFS::AFS::Util->remove_mount(
            volume => qq{efs.$type},
            parent => q{efs.root},
            path   => $type,
            cell   => $cell->name,
        );
        EFS::AFS::Util->remove_volume(
            name => qq{efs.$type},
            cell => $cell->name,
        );
    }    

    EFS::AFS::Util->remove_mount(
        volume => q{efs.root},
        parent => q{root.cell},
        path   => q{efs},
        cell   => $cell->name,
    );

    EFS::AFS::Util->remove_volume(
        name => q{efs.root},
        cell => $cell->name,
    );

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
