package EFS::CLI::Fileshare;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};
use Params::Validate qw( :all );

sub showexports {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        name       => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    $cell->assert_fstype( q{nfs} );

    my $resultset = $class->get_resultset;

    my $fileshare = $resultset->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        name       => $args{name},
    );

    my @efsservers = $cell->efsservers;

    my $efsservers = join q{:}, map { $_->name } @efsservers;

    my @exports = qw( sec=sys );

    if ( $fileshare->filetype eq q{dev} ) {
        push @exports, q{rw};
    } else {
        push @exports, q{ro}, qq{rw=$efsservers};
    }

    push @exports, qq{root=$efsservers};

    STDOUT->print(
        qq{Export definition for $fileshare:\n},
        join( q{,}, @exports ), qq{\n},
    );

    return 1;

}

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        name       => { type => SCALAR },
        filetype   => { type => SCALAR, optional => 1, default => q{none} },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $fileserver_rs = $class->get_resultset( q{Fileserver} );

    my $fileserver = $fileserver_rs->retrieve_bynames_assert(
        cell => $cell->name,
        name => $args{fileserver},
    );

    my $fileserver_name = $fileserver->name;

    if ( $cell->fstype eq q{afs} ) {
        if ( $args{filetype} ne q{none} ) {
            die qq{Invalid argument: --filetype is 'none' for afs cells\n};
        }
    } else {
        if ( $args{filetype} ne q{dist} && $cell->celltype eq q{prod} ) {
            die(
                qq{Fileshares of filetype '$args{filetype}' can only be },
                qq{defined in 'dev' cells.\n},
                qq{Cell '$cell' has celltype 'prod'.\n}
            );
        }
    }

    my $resultset = $class->get_resultset;

    my ($fileshare) = $resultset->search({
        fileserver => $fileserver->id,
        name       => $args{name},
    });

    if ( $fileshare ) {

        if ( $fileshare->filetype eq $args{filetype} ) {
            STDOUT->print(
                q{Verified definition of fileshare },
                qq{'$cell/$fileserver_name/$args{name}'.\n}
            );
        } else {
            $fileshare->set_column( filetype => $args{filetype} );
            $fileshare->update();
            STDOUT->print(
                q{Changed filetype definition of fileshare },
                qq{'$cell/$fileserver_name/$args{name}' },
                qq{to '$args{filetype}'\n}
            );
        }

    } else {

        $fileshare = $resultset->create({
            fileserver => $fileserver->id,
            name       => $args{name},
            filetype   => $args{filetype},
        });

        STDOUT->print( qq{Defined fileshare '$fileshare', filetype '$args{filetype}'\n} );

    }

    return $fileshare;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        name       => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $fileshare = $resultset->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        name       => $args{name},
    );

    if ( $fileshare->cell->fstype eq q{nfs} ) {
        if ( my @fileqtrees = $fileshare->fileqtrees ) {
            die(
                qq{Unable to undef fileshare '$fileshare', },
                qq{as it has the following fileqtrees:\n\t},
                join(qq{\n\t}, map { $_->name } @fileqtrees ), qq{\n},
            );
        }
    }

    my $object_id = $fileshare->object_id;

    $fileshare->attrs->delete_all;
    $fileshare->delete;

    STDOUT->print( qq{Undefined fileshare '$object_id'\n} );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type     => SCALAR,
                        optional => 1 },
        fileserver => { type     => SCALAR,
                        depends  => [qw( cell )],
                        optional => 1 },
        name       => { type     => SCALAR,
                        depends  => [qw( cell fileserver )],
                        optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my @fileshares = ();

    if ( not $args{cell} ) {

        @fileshares = $resultset->all;

    } elsif ( not $args{fileserver} ) {

        my $cell_rs = $class->get_resultset( q{Cell} );
        my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
        @fileshares = $cell->fileshares;

    } elsif ( not $args{name} ) {

        my $fileserver_rs = $class->get_resultset( q{Fileserver} );

        my $fileserver = $fileserver_rs->retrieve_bynames_assert(
            cell => $args{cell},
            name => $args{fileserver},
        );

        @fileshares = $fileserver->fileshares;

    } else {

        $fileshares[0] = $resultset->retrieve_bynames_assert(
            cell       => $args{cell},
            fileserver => $args{fileserver},
            name       => $args{name},
        );

    }

    $class->display_objects( objects => \@fileshares );

    if ( scalar @fileshares != 1 ) {
        return 1;
    }

    my $fileshare = $fileshares[0];

    if ( my @attrs = $fileshare->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( my @fileqtrees = $fileshare->fileqtrees ) {
        $fileqtrees[0]->display_objects(
            objects         => \@fileqtrees,
            display_columns => [qw( fileshare name )],
        );
    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        name       => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{fileshare} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

sub execlinks {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        name       => { type => SCALAR },
        force      => { type => SCALAR | UNDEF, optional => 1 },
    });

    $class->assert_maintenance_mode;

    my $resultset = $class->get_resultset;

    my $fileshare = $resultset->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        name       => $args{name},
    );

    $fileshare->cell->assert_fstype( q{nfs} );

    return EFS::RPC->execute(
        cell   => $fileshare->cell,
        class  => $class,
        method => q{execlinks_rpc},
        args   => [ %args ],
    );

}

sub execlinks_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $fileshare = $resultset->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        name       => $args{name},
    );

    my ($fileqtree) = $fileshare->fileqtrees;

    EFS::CLI::Fileqtree->execlinks_rpc(
        cell       => $fileqtree->cell->name,
        fileserver => $fileqtree->fileserver->name,
        fileshare  => $fileqtree->fileshare->name,
        name       => $fileqtree->name,
        force      => $args{force},
    );

    STDOUT->print( qq{Updated execlinks on all fileqtrees\n} );

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
