package EFS::CLI::PTS;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use Params::Validate qw( :all );

sub interactive { return _unsupported_interactive(); }
sub quit        { return _unsupported_interactive(); }
sub sleep       { return _unsupported_interactive(); }
sub source      { return _unsupported_interactive(); }

sub _unsupported_interactive {
    die qq{Interactive pts commands are not supported when using EFS\n};
}

sub apropos { return _unsupported_help(); }
sub help    { return _unsupported_help(); }

sub _unsupported_help {
    die qq{pts help commands are not supported when using EFS\n};
}

# adduser: add a user to a group

sub adduser {

    my $class = shift;

    # Usage: pts adduser -user <user name>+ -group <group name>+
    my (%args) = validate( @_, {
        user  => { type => SCALAR },
        group => { type => SCALAR },
    });

    my $ptsuser_rs  = $class->get_resultset( q{PTSUser} );
    my $ptsgroup_rs = $class->get_resultset( q{PTSGroup} );

    
    
}

# chown: change ownership of a group
#
# Change the ownership in the database, and use that metadata for
# authorization decisions, just as ptserver would.  However, the
# actual object ownership in the real pts database will be
# system:administrators, to prevent users from bypassing EFS and using
# pts directly to alter membership.

sub chown {

    my $class = shift;

    # Usage: pts chown -name <group name> -owner <new owner>
    my (%args) = validate( @_, {
        name  => { type => SCALAR },
        owner => { type => SCALAR },
    });

}

# creategroup     create a new group
# createuser      create a new user
#
# Both of these commands should at least warn (configurably, die) when
# the user ID doesn't match the UNIX ID (if that username is present
# at all).  We need to decide how to handle automatic ID creation, if
# at all.

sub creategroup {

    my $class = shift;

    # Usage: pts creategroup -name <group name>+ [-owner <owner of the group>] [-id <id (negated) for the group>+]
    my (%args) = validate( @_, {
        name  => { type => SCALAR },
        owner => { type => SCALAR, optional => 1 },
        id    => { type => SCALAR, optional => 1 },
    });

}

sub createuser {

    my $class = shift;

    # Usage: pts createuser -name <user name>+ [-id <user id>+]
    my (%args) = validate( @_, {
        name  => { type => SCALAR },
        id    => { type => SCALAR, optional => 1 },
    });

}

# delete          delete a user or group from database

sub delete {

    my $class = shift;

    # Usage: pts delete -nameorid <user or group name or id>+
    my (%args) = validate( @_, {
        nameorid => { type => SCALAR },
    });

}

# removeuser      remove a user from a group

sub removeuser {

    my $class = shift;

    # Usage: pts removeuser -user <user name>+ -group <group name>+
    my (%args) = validate( @_, {
        user  => { type => SCALAR },
        group => { type => SCALAR },
    });

}

# rename          rename user or group

sub rename {

    my $class = shift;

    # Usage: pts rename -oldname <old name> -newname <new name>
    my (%args) = validate( @_, {
        oldname => { type => SCALAR },
        newname => { type => SCALAR },
    });

}

# setfields       set fields for an entry

sub setfields {

    my $class = shift;

    # Usage: pts setfields -nameorid <user or group name or id>+
    #                      [-access <set privacy flags>]
    #                      [-groupquota <set limit on group creation>]
    my (%args) = validate( @_, {
        nameorid   => { type => SCALAR },
        access     => { type => SCALAR, optional => 1 },
        groupquota => { type => SCALAR, optional => 1 },
    });

}

# setmax          set max id
#
# Not at all clear how to handle this one, or even if it's really
# necessary.
sub setmax {

    my $class = shift;

    # Usage: pts setmax [-group <group max>] [-user <user max>]
    my (%args) = validate( @_, {
        group => { type => SCALAR, optional => 1 },
        user  => { type => SCALAR, optional => 1 },
    });

}

# Just pass through and display, after auth checks
#
# examine         examine an entry
# listentries     list users/groups in the protection database
# listmax         list max id
# listowned       list groups owned by an entry or zero id gets orphaned groups
# membership      list membership of a user or group

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
