package EFS::CLI::Release;

use 5.010;
use strict;
use warnings;
use English;
use feature q{switch};

use base q{EFS::CLI};
use base q{EFS::DBI};
use base q{EFS::DistableObject};

use Readonly;
use Params::Validate qw(:all);
use Date::Calc qw(:all);
use Date::Manip;
use File::Basename qw( basename dirname );
use File::Slurp;
use List::MoreUtils qw( uniq );

sub upload {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj    => { type => SCALAR },
        project     => { type => SCALAR },
        name        => { type => SCALAR },
        recurse     => { type => SCALAR, optional => 1, default => 0 },
        force       => { type => SCALAR, optional => 1, default => 0 },
        packageonly => { type => SCALAR, optional => 1, default => 0 },
    });

    my $project_rs = $class->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{upload_rpc},
        args   => [ %args ],
    );

    return 1;

}

# http://trac.openefs.org/ticket/158
#
# We might be able to extend this method to keep working for both releases and
# releasealiases.  The key difference is that releases have dependencies, but
# releasealiases just have the target (an implicit dependency)

sub upload_rpc {

    my $class = shift;
    my %args  = @_;

    my $seen = $args{seen} || {};

    my $resultset = $class->get_resultset;

    # http://trac.openefs.org/ticket/158
    # This must assert that the object is a release
    my $object   = $resultset->retrieve_release_or_alias_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $release = $object;

    if ( $object->isa( q{EFS::DBI::Result::ReleaseAlias} ) ) {
        $release = $object->target_release;
    }

    if ( $args{recurse} ) {

        my $project_rs = $class->get_resultset( q{Project} );

        my $upload_depends = $release->find_mpr_attr( q{upload_depends} ) || q{runtime};
        $upload_depends = q{runtime,build,test} if $upload_depends eq q{all};
        my @upload_depends = split m{,}msx, $upload_depends;

        foreach my $type ( @upload_depends ) {

            my @depends = sort $release->dependencies( type => $type );

            foreach my $depends ( @depends ) {

                my $depends_id = $depends->dependency_object_id;

                next if $seen->{$depends_id}++;

                my ($depends_metaproj,$depends_project,$depends_release) =
                    split( m{/}msx, $depends_id );

                my $project = $project_rs->retrieve_bynames_assert(
                    metaproj => $depends_metaproj,
                    name     => $depends_project,
                );

                EFS::RPC->execute(
                    cell   => $project->cell,
                    class  => $class,
                    method => q{upload_rpc},
                    args   => [
                        history     => $args{history},
                        metaproj    => $depends_metaproj,
                        project     => $depends_project,
                        name        => $depends_release,
                        recurse     => 1,
                        force       => $args{force},
                        seen        => $seen,
                        packageonly => $args{packageonly},
                    ],
                );

            }

        }

    }

    my $package = EFS::Package->new(
        history   => $args{history},
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        force     => $args{force},
    );

    $package->package;

    if ( not $args{packageonly} ) {
        $package->upload;
    }

    return 1;

}

sub download {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR },
        project   => { type => SCALAR },
        name      => { type => SCALAR },
        norecurse => { type => SCALAR, optional => 1, default => 0 },
        force     => { type => SCALAR, optional => 1, default => 0 },
        %EFS::CLI::validate_download_args,
    });

    my $project_rs = $class->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    if ( $project ) {
        EFS::RPC->execute(
            cell   => $project->cell,
            class  => $class,
            method => q{download_rpc},
            args   => [ %args ],
        );
    }
    else {
        $class->download_rpc( %args );
    }

    return 1;

}

sub download_rpc {

    my $class = shift;
    my %args  = @_;

    my @cells = EFS::DBI->get_localcell;

    my $seen = $args{seen} || {};

    my $package = EFS::Package->new(
        history   => $args{history},
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        force     => $args{force},
        cells     => [@cells],
    );

    $package->get_metadata();

    my $download_depends = $package->download_depends;

    if ( not $args{norecurse} ) {

        my $metaproj_rs = $class->get_resultset( q{Metaproj} );
        my $project_rs  = $class->get_resultset( q{Project} );

        # http://trac.openefs.org/ticket/158
        # If the package is a release, then we need the dependencies.
        # If it's a releasealias, then we need the target.

        foreach my $type ( qw( runtime build test ) ) {

            next if not $download_depends->{$type};

            foreach my $depends ( sort $package->dependencies( $type ) ) {

                next if $seen->{$depends}++;

                my ($depends_metaproj,$depends_project,$depends_release) =
                    split( m{/}msx, $depends );

                my $metaproj = $metaproj_rs->retrieve_bynames_assert(
                    name => $depends_metaproj,
                );

                my ($project) = $project_rs->search({
                    metaproj => $metaproj->id,
                    name     => $depends_project,
                });

                my %rpc_args = (
                    history  => $args{history},
                    metaproj => $depends_metaproj,
                    project  => $depends_project,
                    name     => $depends_release,
                    force    => $args{force},
                    seen     => $seen,
                    cells    => [@cells],
                );

                if ( $project ) {
                    EFS::RPC->execute(
                        cell   => $project->cell,
                        class  => $class,
                        method => q{download_rpc},
                        args   => [ %rpc_args ],
                    );
                } else {
                    $class->download_rpc( %rpc_args );
                }

            }

        }

    }

    $package->download;

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type     => SCALAR },
        project  => { type     => SCALAR },
        name     => { type     => SCALAR },
        sortby   => { type     => SCALAR,
                      default  => q{name},
                      optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        destroyed => 1,
    );

    if ( my $destroyed = $release->destroyed ) {
        STDOUT->print(
            qq{NOTICE: Release '$release' was destroyed on '$destroyed'\n}
        );
    }

    $class->display_objects( objects => [ $release ] );

    foreach my $column ( qw( packaged uploaded downloaded ) ) {
        if ( my $date = $release->$column ) {
            my $label = ucfirst $column;
            STDOUT->print( qq{$label on $date\n} );
        }
    }

    if ( my $clone = $release->cloneof ) {
        STDOUT->print(qq{\nThis release is a clone of $clone\n\n});
    }

    if ( my @clones = $release->clones ) {
        STDOUT->print(
            qq{The following releases are clones of this one:\n\t},
            join(qq{\n\t}, sort @clones ), qq{\n},
        );
    }

    my @optional = ();

    my $feature_rs = $class->get_resultset( q{Feature} );

    if ( not $feature_rs->configured( q{disable_restricted_access} ) ) {
        push @optional, $release->get_efs_column_group( q{access} );
    }

    if ( @optional ) {
        $release->display_options( columns => \@optional );
    }

    if ( my @attrs = $release->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( my @installs = $release->installs ) {
        $installs[0]->display_objects(
            objects     => \@installs,
            sort_column => $args{sortby},
        );
    }

    my @releasealiases = $release->all_releasealiases;

    if ( @releasealiases ) {
        $releasealiases[0]->display_objects( objects => \@releasealiases );
    }

    if ( my @dependencies = $release->dependencies ) {
        $dependencies[0]->display_objects(
            title   => qq{Dependencies of '$release'},
            objects => \@dependencies,
        );
    }

    if ( my @dependents = $release->dependents ) {
        $dependents[0]->display_objects(
            title   => qq{Direct Dependents on '$release'},
            objects => \@dependents,
        );
    }

    if ( my @dependents = map { $_->dependents } @releasealiases ) {
        $dependents[0]->display_objects(
            title   => qq{ReleaseAlias Dependents on '$release'},
            objects => \@dependents,
        );
    }

    return 1;

}

sub create {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj        => { type => SCALAR },
        project         => { type => SCALAR },
        name            => { type => SCALAR },
        nocreate        => { type => SCALAR, optional => 1 },
        clonesrc        => { type => SCALAR, optional => 1 },
        _project_locked => { type => SCALAR, optional => 1 },
    });

    my $project_rs = $class->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{create_rpc},
        args   => [ %args ],
    );

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return $release;

}

sub create_rpc {

    my $class = shift;
    my %args  = @_;

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project_rs = $class->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    my $project_dir = $project->get_path(q{dev});

    my $metaproj = $project->metaproj;

    my $project_lock = $args{_project_locked} || EFS::DirLock->new({
        directory => $project_dir,
        retry     => 1,
        command   => $history->request,
    });

    my $releasealias_rs = $class->get_resultset( q{ReleaseAlias} );

    my ($releasealias) = $releasealias_rs->search({
        project => $project->id,
        name    => $args{name},
    });

    if ( $releasealias ) {
        die(
            qq{Unable to create release '$releasealias', },
            qq{since it already exists as a releasealias.\n}
        );
    }

    my $resultset = $class->get_resultset;

    my ($release) = $resultset->search({
        project => $project->id,
        name    => $args{name},
    });

    if ( $release ) {

        if ( $release->destroyed ) {
            die(
                qq{The release '$release' has been destroyed.\n},
                qq{It must be recovered or purged.\n}
            );
        }

        if ( $release->locked ) {
            die(
                qq{The release '$release' has been locked.\n},
                qq{It must be unlocked before it can be re-created.\n}
            );
        }

    } else {

        $release = $resultset->create({
            project => $project->id,
            name    => $args{name},
        });

    }

    if ( $args{nocreate} ) {
        STDOUT->print( qq{Defined release '$release'\n} );
        return $release;
    }

    my $release_dir = $release->get_path(q{dev});

    my $release_lock;

    if ( -d $release_dir ) {
        $release_lock =  EFS::DirLock->new({
            directory => $release_dir,
            retry     => 1,
            command   => $history->request,
        });
    }

    given ( $project->cell->fstype ) {

        when ( q{afs} ) {
            die qq{Unable to create releases in AFS cells yet\n};
            # XXX: Create and mount the release volume
            # Need to decide what will be a volume and what a plain dir
            # Do we need build volumes, like we had for /ms?
        }

        when ( q{nfs} ) {
            EFS::Util->MakeDir( path => $release_dir );
            EFS::Util->MakeDir( path => qq{$release_dir/install} );
            EFS::Util->MakeDir(
                path  => qq{$release_dir/build},
                user  => $release->get_owner( type => q{user} ),
                group => $release->get_owner( type => q{group} ),
                mode  => $release->get_dirmode,
            );
        }

    }

    EFS::Util->MakeDir(
        path  => qq{$release_dir/build},
        user  => $release->get_owner( type => q{user} ),
        group => $release->get_owner( type => q{group} ),
        mode  => $release->get_dirmode,
    );

    if ( $args{clonesrc} ) {

        my $clone = $resultset->retrieve_bynames_assert(
            metaproj => $args{metaproj},
            project  => $args{project},
            name     => $args{clonesrc},
        );

        my $clone_dir = $clone->get_path( q{dev} );

        EFS::Util->CloneDirectory(
            from    => qq{$clone_dir/src},
            to      => qq{$release_dir/src},
        );

        EFS::CLI::Release->_lock_unlock_rpc(
            history         => $args{history},
            metaproj        => $args{metaproj},
            project         => $args{project},
            name            => $args{name},
            lockmode        => q{unlock},
            norecurse       => 1,
            _release_locked => 1,
        );

    }
    else {

        EFS::Util->MakeDir(
            path  => qq{$release_dir/src},
            user  => $release->get_owner( type => q{user} ),
            group => $release->get_owner( type => q{group} ),
            mode  => $release->get_dirmode,
        );

    }

    STDOUT->print( qq{Creating dist2dev symlink for $release\n} );

    given ( $project->cell->fstype ) {

        when ( q{afs} ) {
            die qq{Unable to create dist2dev links in AFS yet\n};
            # XXX: Create the metaproj/project volumes in /efs/dist, and mount
            # them, then create the symlink
        }

        when ( q{nfs} ) {

            my $admin_path = $project->cell->admin_path( type => q{dist} );

            my $dist2dev_dirparent = qq{$admin_path/dist/$metaproj};

            my $dist2dev_dir  = qq{$admin_path/dist/$project};
            my $dist2dev_link = qq{$admin_path/dist/$release};
            my $dist2dev_show = qq{/efs/dist/$release};

            if ( -d $dist2dev_link && ! -l $dist2dev_link ) {
                die(
                    qq{Unable to create dist2dev link ($dist2dev_show)\n},
                    qq{It already exists as real directory\n}
                );
            } else {
                EFS::Util->MakeDir( path => $dist2dev_dirparent );
                EFS::Util->MakeDir( path => $dist2dev_dir );
                EFS::Util->MakeLink(
                    from => $dist2dev_link,
                    to   => qq{/efs/dev/$release/install},
                );
            }
    
            STDOUT->print( qq{_EFS_WAITFOR: $release_dir\n} );
            STDOUT->print( qq{_EFS_WAITFOR: $dist2dev_show\n} );

        }

    }

    STDOUT->print( qq{Created release $release\n} );

    return $release;

}

sub chown {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR | OBJECT },
        project   => { type => SCALAR | OBJECT },
        name      => { type => SCALAR | OBJECT },
        username  => { type => SCALAR, optional => 1 },
        norecurse => { type => BOOLEAN, optional => 1 },
        verbose   => { type => BOOLEAN, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{chown_rpc},
        args   => [ %args ],
    );

}

sub chown_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project  = $release->project;
    my $metaproj = $project->metaproj;

    my $release_dir = qq{/efs/dev/$release};

    my $groupname = $release->get_owner( type => q{group} );
    my $username  = $release->get_owner( type => q{user} );

    if ( $args{username} and $args{username} ne $username ) {

        my $username_obj = EFS::Username->new({ name => $args{username} });

        if ( not $username_obj->is_member_of( $groupname ) ) {
            die(
                qq{Invalid username '$args{username}'.\n},
                qq{Must be a member of the group '$groupname'\n},
            );
        }

        $username = $args{username};

    }

    my $lockgroup = $release->get_owner(
        lockmode => q{lock},
        type     => q{group},
    );

    my $dirlock = EFS::DirLock->new({
        directory => $release_dir,
        retry     => 1,
        command   => $history->request,
    });

    foreach my $subdir ( qw( src build ) ) {
        STDOUT->print(qq{Chowning directory $release_dir/$subdir\n});
        EFS::Util->Chown(
            path          => qq{$release_dir/$subdir},
            user          => $username,
            group         => $groupname,
            lockgroup     => $lockgroup,
            verbose       => $args{verbose},
            noworldaccess => $release->get_noworldaccess,
        );
    }

    if ( $args{norecurse} ) {
        return 1;
    }

    my @installs = sort $release->installs;

    if ( not @installs ) {
        STDOUT->print( qq{No installs yet created for release '$release'\n} );
        return 1;
    }

    STDOUT->print( qq{Chowning all installs in release '$release'\n} );

    my $fm_installs = EFS::ForkManager->new({

        list => \@installs,

        action => sub {

            my $install = shift;

            EFS::CLI::Install->chown(
                history         => $args{history},
                metaproj        => $args{metaproj},
                project         => $args{project},
                release         => $args{name},
                name            => $install->name,
                username        => $username,
                verbose         => $args{verbose},
                _release_locked => 1,
            );

        },

    });

    $fm_installs->run || die(
        qq{Unable to chown the following installs:\n\t},
        join( qq{\n\t}, $fm_installs->get_failed ), qq{\n}
    );

    STDOUT->print( qq{Successfully chowned all installs in release '$release'\n} );

    return 1;

}

sub lock {
    return shift->_lock_unlock( @_, lockmode => q{lock} );
}

sub unlock {
    return shift->_lock_unlock( @_, lockmode => q{unlock} );
}

sub _lock_unlock {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj        => { type     => SCALAR },
        project         => { type     => SCALAR },
        name            => { type     => SCALAR },
        lockmode        => { type     => SCALAR },
        username        => { type     => SCALAR, optional => 1 },
        norecurse       => { type     => BOOLEAN, optional => 1 },
        verbose         => { type     => BOOLEAN, optional => 1 },
        _release_locked => { type     => BOOLEAN, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{_lock_unlock_rpc},
        args   => [ %args ],
    );

}

sub _lock_unlock_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $release_dir = $release->get_path(q{dev});

    my $mode_verb = $args{lockmode} eq q{lock} ? q{Locking} : q{Unlocking};

    my $feature_rs = $class->get_resultset( q{Feature} );

    if ( $feature_rs->configured( q{allow_insecure_disable_lockdown} ) ) {
        if ( $release->find_mpr_attr( q{disable_lockdown} ) ) {
            die(
                qq{Release '$release' has the attribute 'disable_lockdown',\n},
                qq{and can't be locked, explicitly or implicitly.\n},
            );
        }
    }

    my $username = $release->get_owner(
        lockmode => $args{lockmode},
        type     => q{user},
    );

    my $groupname = $release->get_owner(
        lockmode => $args{lockmode},
        type     => q{group},
    );

    if ( $args{lockmode} eq q{unlock} ) {

        if ( $args{username} and $args{username} ne $username ) {

            my $username_obj = EFS::Username->new({ name => $args{username} });

            if ( not $username_obj->is_member_of( $groupname ) ) {
                die(
                    qq{Invalid username '$args{username}'.\n},
                    qq{Must be a member of the group '$groupname'\n},
                );
            }

            $username = $args{username};

        }

    }

    my $lockgroup = $release->get_owner(
        lockmode => q{lock},
        type     => q{group},
    );

    my $release_lock = $args{_release_locked} || EFS::DirLock->new({
        directory => $release_dir,
        retry     => 1,
        command   => $history->request,
    });

    my $locked = $args{lockmode} eq q{lock} ? 1 : 0;

    if ( not $locked ) {
        $release->set_column( downloaded => undef );
        $release->set_column( packaged   => undef );
        $release->set_column( uploaded   => undef );
    }
    $release->set_column( locked => $locked );
    $release->update;

    foreach my $subdir ( qw( src build ) ) {

        STDOUT->print(qq{$mode_verb directory $release_dir/$subdir\n});

        EFS::Util->Chown(
            path          => qq{$release_dir/$subdir},
            user          => $username,
            group         => $groupname,
            lockgroup     => $lockgroup,
            verbose       => $args{verbose},
            lockmode      => $args{lockmode},
            noworldaccess => $release->find_mpr_attr( q{noworldaccess} ),
        );

    }

    if ( $args{norecurse} ) {
        return 1;
    }

    my @installs = $release->installs;

    if ( not @installs ) {
        STDOUT->print( qq{No installs yet created for release '$release'\n} );
        return 1;
    }

    if ( $args{lockmode} eq q{unlock} ) {
        foreach my $clone ( $release->clones ) {
            $clone->unclone;
            STDOUT->print( qq{Release $clone is no longer a clone\n} );
        }
    }

    STDOUT->print( qq{$mode_verb all installs in release '$release'\n} );

    my $fm_installs = EFS::ForkManager->new({

        list   => \@installs,

        action => sub {

            my $install = shift;

            EFS::CLI::Install->_lock_unlock(
                history         => $args{history},
                lockmode        => $args{lockmode},
                metaproj        => $args{metaproj},
                project         => $args{project},
                release         => $args{name},
                name            => $install->name,
                username        => $username,
                verbose         => $args{verbose},
                _release_locked => 1,
            );

        },

    });

    my $success = $fm_installs->run;

    # We want the trigger to run if any installs were modified,
    # even if the entire operation was not successful
    if ( $success or $fm_installs->get_succeeded ) {
        $release->_run_lock_unlock_trigger( %args );
    }

    if ( not $success ) {
        die(
            qq{Unable to $args{lockmode} the following installs:\n\t},
            join( qq{\n\t}, $fm_installs->get_failed ), qq{\n}
        );
    }

    STDOUT->print( qq{Successfully chowned all installs in $release\n} );

    return 1;

}

sub checkpoint {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type     => SCALAR | OBJECT },
        project  => { type     => SCALAR | OBJECT },
        name     => { type     => SCALAR | OBJECT },
        query    => { type     => SCALAR,
                      optional => 1 },
        nolock   => { type     => BOOLEAN,
                      optional => 1 },
        noaudit  => { type     => BOOLEAN,
                      optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{checkpoint_rpc},
        args   => [ %args ],
    );

}

sub checkpoint_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $release_dir = $release->get_path(q{dev});

    if ( $args{noaudit} && $release->has_only_prod_stages ) {
        die(
            q{Audits may not be disabled for a release },
            qq{which has a stage of prod or deprecated.\n}
        );
    }

    if ( $release->name eq q{incr} ) {
        $args{nolock}++;
    }

    my $feature_rs = $class->get_resultset( q{Feature} );

    if ( $feature_rs->configured( q{allow_insecure_disable_lockdown} ) ) {
        if ( $release->find_mpr_attr( q{disable_lockdown} ) ) {
            $args{nolock}++;
        }
    }

    my @installs = sort $release->installs;

    if ( not @installs ) {
        die(qq{No installs have been created for release '$release'\n});
    }

    my $release_lock = EFS::DirLock->new({
        directory => $release_dir,
        retry     => 1,
        command   => $history->request,
    });

    my $fm_installs = EFS::ForkManager->new({

        list => \@installs,

        action => sub {
            shift->update_checkpoint(
                history => $args{history},
                nolock  => $args{nolock},
                noaudit => $args{noaudit},
            );
        },

    });

    $fm_installs->run || die(
        qq{Unable to checkpoint the following installs:\n\t},
        join(qq{\n\t}, $fm_installs->get_failed ), qq{\n}
    );

    if ( not $args{nolock} ) {
        $class->lock(
            history         => $args{history},
            metaproj        => $args{metaproj},
            project         => $args{project},
            name            => $args{name},
            norecurse       => 1,
            _release_locked => 1,
        );
    }

    STDOUT->print( qq{Successfully completed checkpoint for '$release'\n} );

    return 1;

}

sub disttree {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
        %EFS::CLI::validate_dist_args,
    });

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
    );

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $tree = EFS::DependencyTree->new({
        seeds => [ $release ],
        type  => q{runtime},
    });

    delete $args{metaproj};
    delete $args{project};
    delete $args{name};

    if ( $args{dryrun} ) {
        STDOUT->print( $tree->get_tree_text( cells => \@cells ) );
        return 1;
    } else {
        return $tree->dist( %args );
    }

}

sub dist {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj      => { type     => SCALAR | OBJECT },
        project       => { type     => SCALAR | OBJECT },
        name          => { type     => SCALAR | OBJECT },
        verbose       => { type     => BOOLEAN, optional => 1 },
        noclone       => { type     => BOOLEAN, optional => 1 },
        _waitfor_cell => { type     => SCALAR, optional => 1 },
        q{no-strict-releasealiases}
                      => { type     => SCALAR, optional => 1 },
        %EFS::CLI::validate_dist_args,
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{dist_rpc},
        args   => [ %args ],
    );

}

sub dist_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    # NOTE: verbose by default, until we print summary stats
    $args{verbose}++;

    my $release_dir   = $release->get_path(q{dev});
    my $release_stage = $release->stage;

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
    );

    if ( grep { $_->celltype eq q{prod} } @cells ) {
        if ( not $release->has_only_prod_stages ) {
            die(
                q{Only prod or deprecated releases may be },
                qq{disted to prod cells.\n},
                qq{Release '$release' is in stage '$release_stage'.\n},
            );
        }
    }

    my $diststate = $release->get_diststate;

    if ( not $diststate->needs_update( cells => \@cells ) ) {
        STDOUT->print( qq{Release $release is up to date in all requested cells.\n} );
        return 1;
    }

    my @dependencies = $release->dependencies( type => q{runtime} );

    my @undisted = ();
    my @partials = ();

    foreach my $dependency ( @dependencies ) {

        my $diststate       = $dependency->get_diststate;
        my $dependency_mpr  = $dependency->dependency_object_id;
        my $dependency_type =
            $dependency->isa(q{EFS::DBI::Result::DependencyRelease}) ?
                q{release} : q{releasealias};

        if ( not $diststate->cells_partially_updated ) {
            STDOUT->print( qq{Dependency $dependency_mpr has not yet been disted.\n} );
            push @undisted, $dependency_mpr;
        }

        if ( $diststate->cells_needing_update( cells => \@cells ) ) {

            my $showstate = 1;

            if ( $dependency_type eq q{release} ) {
                push @undisted, $dependency_mpr;
            } else {
                if ( $diststate->cells_never_updated( cells => \@cells ) ) {
                    push @undisted, $dependency_mpr;
                } else {
                    push @partials, $dependency_mpr;
                    if ( $args{q{no-strict-releasealiases}} ) {
                        $showstate = 0;
                    }
                }
            }

            if ( $showstate ) {
                STDOUT->print(
                    qq{Diststate for $dependency_type },
                    qq{dependency $dependency_mpr:\n}
                );
                $diststate->display_text_table( cells => \@cells );
            }

        }

    }

    if ( @undisted ) {
        die(
            qq{Unable to dist release because the following dependencies\n},
            qq{are not disted to all requested cells.  They must be disted to\n},
            qq{those cells before proceeding any further:\n\t},
            join( qq{\n\t}, @undisted ), qq{\n},
        );
    }

    if ( @partials and not $args{q{no-strict-releasealiases}} ) {
        die(
            qq{Unable to dist release because the following releasealias\n},
            qq{dependencies are not consistently deployed, and have different\n},
            qq{values in different cells.\n},
            qq{If this is acceptable, then you can dist it by using\n},
            qq{the --no-strict-releasealiases argument (--nsra is an alias):\n\t},
            join( qq{\n\t}, @partials ), qq{\n},
        );
    }

    # Cloned release support
    if ( my $clone = $release->cloneof ) {
        if ( $clone->get_diststate->cells_needing_update( cells => \@cells ) ) {
            if ( not $args{noclone} ) {
                die(
                    qq{Clone $clone has not been disted to all cells.\n},
                    qq{To dist without clone optimizations, use --noclone\n},
                );
            }
        }
    }

    # NOTE: We already handle the uptodate case above
    $diststate->display_text_table( cells => \@cells );

    return 1 if $args{dryrun};

    my $dirlock = EFS::DirLock->new({
        directory => $release_dir,
        retry     => 1,
        command   => $history->request,
    });

    $release->validate_checkpoint;

    # cells2installs let's us track which cells need what installs
    # This is what drives the first forkmanager below

    # Well, for hierarchical distribution, we'll need to change that.
    # The forkmanager has to be driven by a list of the dev cells, and
    # when each dev cell complete, it can kick off a dist to its
    # children (prod and/or test cells)

    my %cells2installs  = ();

    foreach my $cell ( @cells ) {
        if ( my @needing_update =
                 $diststate->installs_needing_update( cells => [ $cell ] ) ) {
            $cells2installs{ $cell->name } = [
                map { $_->name } @needing_update
            ];
        }
    }

    if ( not keys %cells2installs ) {
        # We shouldn't get here.  This code used to check that there was still
        # something left in that hash after we failed to make some attaches
        # (an old Acopia concept... RIP), which are a thing of the past.
        die qq{Unable to dist release '$release'\n};
    }

    # If any of the prod (or test) cells have installs to dist, but
    # their parentcell does NOT, then add an empty install list for
    # the parentcell. This will allow us to loop over the dev cells in
    # @cells to drive the first forkmanager.  If the dev cell's NOT in
    # this data structure, then either none of its child cells have
    # dists pending.

    foreach my $cell ( @cells ) {
        if ( $cell->celltype ~~ [qw( test prod )]
                 and $cells2installs{ $cell->name }
                     and not $cells2installs{ $cell->parentcell->name } ) {
            $cells2installs{ $cell->parentcell->name } = [];
        }
    }

    #
    # We need to remove the dist2dev symlink, if it exists.
    #
    if ( -l qq{/efs/dist/$release} ) {
        my $admin_path = $release->cell->admin_path( type => q{dist} );
        EFS::Util->RemoveLink( from => qq{$admin_path/dist/$release} );
    }

    my @dev_cells = grep { $_->celltype eq q{dev} &&
                               $cells2installs{ $_->name } } @cells;

    my $fm_cells = EFS::ForkManager->new({

        # See above -- we'll need to drive this forkmanager off of a
        # list of the dev cells only.

        list => \@dev_cells,

        action => sub {

            my $cell    = shift;

            if ( my @install_names = @{ $cells2installs{ $cell->name } } ) {

                # Don't use EFS::RPC here.  The first pass at disting
                # something has to be done from the project's cell.
                # Disting from dev to prod/test cells can use EFS::RPC

                $class->dist_cell(
                    history   => $history->id,
                    src_cell  => $release->cell->name,
                    dest_cell => $cell->name,
                    metaproj  => $release->metaproj->name,
                    project   => $release->project->name,
                    release   => $release->name,
                    installs  => \@install_names,
                );

            }

            my @child_cells     = ();

            foreach my $child_cell ( $cell->children ) {
                if ( $cells2installs{ $child_cell->name } ) {
                    push( @child_cells, $child_cell );
                }
            }

            if ( not @child_cells ) {
                return 1;
            }

            my $fm_children = EFS::ForkManager->new({

                list => \@child_cells,

                action => sub {

                    my $child = shift;

                    my @install_names =
                        @{ $cells2installs{ $child->name } };

                    return EFS::RPC->execute(
                        cell    => $cell,
                        class   => $class,
                        method  => q{dist_cell},
                        args    => [
                            history   => $history->id,
                            src_cell  => $cell->name,
                            dest_cell => $child->name,
                            metaproj  => $release->metaproj->name,
                            project   => $release->project->name,
                            release   => $release->name,
                            installs  => \@install_names,
                        ],
                    );

                },

            });

            $fm_children->run || die(
                qq{Unable to distribute from $cell },
                qq{to the following child cells:\n\t},
                join( qq{\n\t}, $fm_children->get_failed ), qq{\n},
            );

            return 1;

        },

    });

    # XXX: Display dist stats here...

    $fm_cells->run || die(
        q{Unable to distribute to the following cells },
        qq{(or possibly their children):\n\t},
        join(qq{\n\t}, $fm_cells->get_failed ) . qq{\n}
    );

    STDOUT->print( qq{Successfully distributed release to all cells\n} );

    if ( $release->cloneof and
             $release->has_prod_stages and
                 not $release->get_diststate->cells_needing_update ) {
        $release->unclone;
        STDOUT->print(
            qq{Release $release is no longer a clone (globally updated in prod)\n}
        );
    }

    if ( $args{_waitfor_cell} ) {
        my $release_dir = $release->get_path(q{dist});
        if ( grep { $args{_waitfor_cell} eq $_ } $fm_cells->get_succeeded ) {
            STDOUT->print( qq{_EFS_WAITFOR: $release_dir\n} );
        }
    }

    return 1;

}

sub dist_cell {

    my $class = shift;

    my (%args) = validate( @_, {
        history   => { type => SCALAR },
        src_cell  => { type => SCALAR },
        dest_cell => { type => SCALAR },
        metaproj  => { type => SCALAR },
        project   => { type => SCALAR },
        release   => { type => SCALAR },
        installs  => { type => ARRAYREF },
    });

    my $history_rs = $class->get_resultset( q{History} );
    my $history    = $history_rs->retrieve_assert( id => $args{history} );

    my $cell_rs   = $class->get_resultset( q{Cell} );
    my $src_cell  = $cell_rs->retrieve_bynames_assert( name => $args{src_cell} );
    my $dest_cell = $cell_rs->retrieve_bynames_assert( name => $args{dest_cell} );

    my $resultset = $class->get_resultset;
    my $release   = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{release},
    );
    my $release_mpr = $release->object_id;

    my $disable_exec_backwards = $release->find_mpr_attr(q{disable_exec_backwards}),
    my $disable_exec_emulates  = $release->find_mpr_attr(q{disable_exec_emulates}),

    STDOUT->print( qq{Distributing $release from $src_cell to $dest_cell\n} );

    my $starttime = time;

    my $diststats_rs = $class->get_resultset( q{DistStats} );

    my ($diststats) = $diststats_rs->create({
        history   => $history->id,
        metaproj  => $release->metaproj->name,
        project   => $release->project->name,
        release   => $release->name,
        src_cell  => $src_cell->name,
        dest_cell => $dest_cell->name,
    });

    my $diststats_id = $diststats->id;

    my $admin_path = $dest_cell->admin_path( type => q{dist} );
    my $dest_path  = qq{$admin_path/dist/$release};

    my %installs = map { $_->name => $_ } $release->installs;

    my $dest_path_bottom = $dest_path;

    if ( grep { $_ ne q{common} } keys %installs ) {
        $dest_path_bottom .= q{/.exec};
    }

    EFS::RPC->execute(
        cell   => $dest_cell,
        class  => q{EFS::Util},
        method => q{MakeDirHierarchy},
        args   => [
            top    => $admin_path,
            bottom => $dest_path_bottom,
        ],
    );

    my $fm_installs = EFS::ForkManager->new({

        list => $args{installs},

        action => sub {

            my $name = shift;

            my $install = $installs{ $name };

            $install->distribute(
                diststats_id => $diststats_id,
                src_cell     => $src_cell,
                dest_cell    => $dest_cell,
                verbose      => $args{verbose},
            );

            return 1;

        },

    });

    my $status = $fm_installs->run;

    my $endtime = time;

    $diststats->status( $status );
    $diststats->rollup_rsyncstats( $endtime - $starttime );
    $diststats->display_stats if $args{verbose};
    $diststats->update;

    #
    # NOTE: Create these links if *any* of the installs were
    # successfully updated.
    #
    if ( my @successful = $fm_installs->get_succeeded ) {

        # Don't bother with this unless something other than common
        # was updated.

        if ( grep { $_ ne q{common} } @successful ) {

            EFS::RPC->execute(
                cell    => $dest_cell,
                class   => q{EFS::ExecLinks},
                method  => q{update_links},
                args    => [
                    command                => $history->request,
                    directory              => qq{$dest_path/.exec},
                    disable_exec_backwards => $disable_exec_backwards,
                    disable_exec_emulates  => $disable_exec_emulates,
                ],
            );

        }

        my $cellstate_rs = $class->get_resultset( q{InstallCellState} );

        foreach my $name ( @successful ) {

            my $install = $installs{ $name };

            $cellstate_rs->update_or_insert(
                install    => $install,
                cell       => $dest_cell,
                updatetime => $install->checktime,
            );

        }

        STDOUT->print(
            q{Successfully distributed the following },
            qq{installs to cell $dest_cell:\n\t},
            join( qq{\n\t}, map { qq{$release_mpr/$_} } @successful ) . qq{\n},
        );

    }

    if ( not $status ) {
        die(
            q{Unable to distribute the following },
            qq{installs to cell $dest_cell:\n\t},
            join( qq{\n\t}, $fm_installs->get_failed ) . qq{\n},
        );
    }

    STDOUT->print( qq{Successfully distributed all installs to $dest_cell\n} );

    $src_cell->wait_for_dist_cache;

    return 1;

}

sub execlinks {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR | OBJECT },
        project  => { type => SCALAR | OBJECT },
        name     => { type => SCALAR | OBJECT },
        %EFS::CLI::validate_diststate_args,
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{execlinks_rpc},
        args   => [ %args ],
    );

}

sub execlinks_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $disable_exec_backwards = $release->find_mpr_attr(q{disable_exec_backwards});
    my $disable_exec_emulates  = $release->find_mpr_attr(q{disable_exec_emulates});

    my $locked = $release->locked;

    my $history_rs = $class->get_resultset( q{History} );
    my $history    = $history_rs->retrieve_assert( id => $args{history} );

    my $diststate = $release->get_diststate;

    my @cells = $class->handle_cell_arguments(
        %args,
        default_scope => q{global},
    );

    my @disted_cells =
        grep { not $_->hidden }
            $diststate->cells_partially_updated( cells => \@cells );

    my @installs = $release->installs;

    if ( not @installs ) {
        STDOUT->print( qq{Release '$release' has no installs.\n} );
        return 1;
    }

    if ( not grep { $_->name ne q{common} } @installs ) {
        STDOUT->print( qq{Release '$release' has no exec installs\n} );
        return 1;
    }

    my $admin_path = $release->cell->admin_path( type => q{dev} );

    EFS::ExecLinks->update_links(
        command                => $history->request,
        directory              => qq{$admin_path/dev/$release/install/.exec},
        disable_exec_backwards => $disable_exec_backwards,
        disable_exec_emulates  => $disable_exec_emulates,
        locked                 => $locked,
    );

    STDOUT->print( qq{Updated execlinks for '$release' in /efs/dev\n} );

    if ( not @disted_cells ) {
        STDOUT->print( qq{Nothing to update in /efs/dist\n} );
        return 1;
    }

    my $fm_cells = EFS::ForkManager->new({

        list => \@disted_cells,

        action => sub {

            my $cell = shift;
            my $admin_path = $cell->admin_path( type => q{dist} );
            my $exec_path = qq{$admin_path/dist/$release/.exec};

            return EFS::RPC->execute(
                cell    => $cell,
                class   => q{EFS::ExecLinks},
                method  => q{update_links},
                args    => [
                    command                => $history->request,
                    directory              => $exec_path,
                    disable_exec_backwards => $disable_exec_backwards,
                    disable_exec_emulates  => $disable_exec_emulates,
                    locked                 => $locked,
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to update execlinks in the following cells:\n\t},
        join(qq{\n\t}, $fm_cells->get_failed ) . qq{\n}
    );

    STDOUT->print( qq{Updated execlinks for '$release' in /efs/dist\n} );

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{release} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

sub destroy {

    my $class = shift;

    my %args = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj        => { type     => SCALAR },
        project         => { type     => SCALAR },
        name            => { type     => SCALAR },
        force           => { type     => SCALAR, optional => 1 },
        _project_locked => { type     => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        destroyed => 1,
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{destroy_rpc},
        args   => [ %args ],
    );

}

sub destroy_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        destroyed => 1,
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $release_stage = $release->stage;
    my $release_dir   = $release->get_path(q{dev});
    my $project_dir   = $release->project->get_path(q{dev});

    if ( not $release->has_only_destroyable_stages ) {
        die(
            qq{Release '$release' is in stage '$release_stage'\n},
            q{Releases must be either rejected or deprecated },
            qq{before they can be destroyed.\n}
        );
    }

    if ( my @releasealiases = $release->releasealiases ) {
        die(
            qq{The following releasealiases must be destroyed or changed before\n},
            qq{this release can be destroyed:\n\t},
            join( qq{\n\t}, @releasealiases), qq{\n},
        );
    }

    if ( $release->dependents || $release->releasealias_dependents ) {
        STDOUT->print(
            qq{The release '$release' has dependents defined.\n},
            qq{Some of these may be direct, and some via releasealiases.\n},
            qq{These must be undefined before this release can be destroyed.\n}
        );
        EFS::CLI::Dependents->display(
            metaproj => $release->metaproj->name,
            project  => $release->project->name,
            release  => $release->name,
        );
        die(
            qq{Unable to destroy release '$release' until these },
            qq{dependencies have been undefined.\n}
        );
    }

    if ( my @clones = $release->clones ) {

        if ( $args{force} ) {
            foreach my $clone ( @clones ) {
                $clone->unclone;
                STDOUT->print( qq{Release $clone is no longer a clone\n} );
            }
        }
        else {
            die(
                qq{Release $release has the following clones:\n\t},
                join( qq{\n\t}, @clones ), qq{\n},
                qq{To destroy this release, and unclone them, use --force\n},
            );
        }

    }

    if ( $release->cloneof ) {
        $release->unclone;
    }

    my $project_lock = $args{_project_locked} || EFS::DirLock->new({
        directory => $project_dir,
        retry     => 1,
        command   => $history->request,
    });

    my $release_lock;

    if ( -d $release_dir ) {
        $release_lock = EFS::DirLock->new({
            directory => $release_dir,
            alternate => qq{$release_dir/.destroyed},
            retry     => 1,
            command   => $history->request,
        });
    }

    my $diststate = $release->get_diststate;

    my @cells = uniq (
        $release->project->cell,        
        grep { not $_->hidden } $diststate->cells_partially_updated
    );

    my $destroyed = $release->project->get_path(q{dev}) . q{/.destroyed};

    if ( $release->cell->fstype eq q{nfs} ) {
        EFS::Util->MakeDir(
            path => $destroyed,
            mode => 0700,
        );
    }

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell        => $cell,
                class       => $class,
                method      => q{destroy_cell},
                args        => [
                    cell     => $cell->name,
                    metaproj => $release->metaproj->name,
                    project  => $release->project->name,
                    release  => $release->name,
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to destroy the release in the following cells:\n\t},
        join( qq{\n\t}, $fm_cells->get_failed ) . qq{\n},
    );

    if ( $release->uploaded ) {
        my $package = EFS::Package->new(
            history => $args{history},
            metaproj  => $args{metaproj},
            project   => $args{project},
            name      => $args{name},
        );
        $package->destroy;
    }

    my $dev_path  = $release->get_path(q{dev});

    given ( $release->cell->fstype ) {

        when ( q{afs} ) {
            die qq{Unable to destroy releases in AFS cells yet\n};
            # XXX: unmount the dev release volume
        }

        when ( q{nfs} ) {
            my $move_to = qq{$destroyed/} . $release->name;
            if ( -d $dev_path ) {
                EFS::Util->RenameDir(
                    from => $dev_path,
                    to   => $move_to,
                );
                EFS::Util->RemoveFile( file => qq{$move_to/.lock} );
            }
        }

    }

    STDOUT->print( qq{Destroyed $dev_path\n} );

    $release->set_column( destroyed => join(q{-}, Today ) );
    $release->update;

    STDOUT->print( qq{Successfully destroyed $release globally\n} );

    return 1;

}

sub destroy_cell {

    my $class = shift;

    my (%args) = validate( @_, {
        cell     => { type => SCALAR },
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        release  => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $resultset = $class->get_resultset;
    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{release},
        destroyed => 1,
    );

    my $project = $release->project;

    my $admin_path = $cell->admin_path( type => q{dist} );

    if ( not -d qq{$admin_path/dist/$release} ) {
        STDOUT->print(
            qq{Release $release was never disted to $cell. Nothing to destroy.\n}
        );
        return 1;
    }

    if ( -l qq{$admin_path/dist/$release} ) {
        unlink( qq{$admin_path/dist/$release} ) ||
            die qq{Unable to unlink $admin_path/dist/$release: $ERRNO\n};
        STDOUT->print( qq{Removed dist2dev symlink $cell:/efs/dist/$release\n} );
    } elsif ( -d qq{$admin_path/dist/$release} ) {

        given ( $cell->fstype ) {

            when ( q{afs} ) {
                die qq{Unable to destroy releases in AFS cells yet\n};
                # XXX: umount the dist release volume
            }

            when ( q{nfs} ) {
                EFS::Util->MakeDir(
                    path => qq{$admin_path/dist/$project/.destroyed},
                    mode => 0700,
                );
                EFS::Util->RenameDir(
                    from => qq{$admin_path/dist/$release},
                    to   => qq{$admin_path/dist/$project/.destroyed/} . $release->name,
                );
            }

        }

        STDOUT->print( qq{Destroyed $release in $cell\n} );
    } else {
        STDOUT->print( qq{Nothing to destroy in cell $cell\n} );
    }

    return 1;

}

sub recover {

    my $class = shift;

    my %args = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type     => SCALAR },
        project  => { type     => SCALAR },
        name     => { type     => SCALAR },
        help     => { type     => BOOLEAN,
                      optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        destroyed => 1,
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{recover_rpc},
        args   => [ %args ],
    );

}

sub recover_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        destroyed => 1,
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project_dir = $release->project->get_path(q{dev});
    my $cell        = $release->cell;

    if ( not $release->destroyed ) {
        die qq{Release '$release' has not been destroyed\n};
    }

    my $dev_path     = $release->get_path(q{dev});
    my $destroyed    = qq{$project_dir/.destroyed};
    my $recover_from = qq{$destroyed/} . $release->name;

    my $project_lock = $args{_project_locked} || EFS::DirLock->new({
        directory => $project_dir,
        retry     => 1,
        command   => $history->request,
    });

    given ( $cell->fstype ) {

        when ( q{afs} ) {
            die qq{Unable to recover releases in AFS cells yet\n};
            # XXX: remount the release volume
        }

        when ( q{nfs} ) {
            if ( -d $recover_from ) {
                if ( -f qq{$recover_from/.lock} ) {
                    unlink( qq{$recover_from/.lock} ) ||
                        die qq{Unable to unlink $recover_from/.lock: $ERRNO\n};
                }

                EFS::Util->RenameDir(
                    from => $recover_from,
                    to   => $dev_path,
                );

                STDOUT->print( qq{Recovered $dev_path\n} );

            } else {
                if ( not -d $dev_path ) {
                    die qq{Unable to find release '$release' in /efs/dev.\n};
                }
                STDOUT->print( qq{Release '$release' has already been recovered in /efs/dev.\n} );
            }
        }

    }

    if ( $release->uploaded ) {
        my $package = EFS::Package->new(
            history   => $args{history},
            metaproj  => $args{metaproj},
            project   => $args{project},
            name      => $args{name},
        );
        $package->recover;
    }

    my $diststate = $release->get_diststate;

    my @cells = grep { not $_->hidden } $diststate->cells_partially_updated;

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell         => $cell,
                class        => $class,
                method       => q{recover_cell},
                args         => [
                    cell     => $cell->name,
                    metaproj => $release->metaproj->name,
                    project  => $release->project->name,
                    release  => $release->name,
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to recover '$release' in the following cells:\n\t},
        join( qq{\n\t}, $fm_cells->get_failed ) . qq{\n},
    );

    $release->set_column( destroyed => undef );
    $release->update;

    STDOUT->print( qq{Successfully recovered '$release' in all cells\n} );

    return 1;

}

sub recover_cell {

    my $class = shift;

    my (%args) = validate( @_, {
        cell     => { type => SCALAR },
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        release  => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $resultset = $class->get_resultset;
    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{release},
        destroyed => 1,
    );

    my $project = $release->project;

    given ( $project->cell->fstype ) {

        when ( q{afs} ) {
            die qq{Unable to recover releases in AFS cells yet\n};
            # XXX: remount the release volume
        }

        when ( q{nfs} ) {

            my $admin_path = $cell->admin_path( type => q{dist} );

            my $to_dir   = qq{$admin_path/dist/$release};
            my $from_dir = qq{$admin_path/dist/$project/.destroyed/} . $release->name;

            if ( not -d $from_dir ) {
                die(
                    qq{Release $release was never disted to $cell, },
                    qq{or has already been recovered here.\n},
                    qq{Nothing to recover.\n}
                );
            }

            if ( -d $to_dir && ! -d $from_dir ) {
                STDOUT->print(
                    qq{Release $release has already been recovered in /efs/dist in $cell\n}
                );

            } elsif ( -d $from_dir ) {
                EFS::Util->RenameDir(
                    from => $from_dir,
                    to   => $to_dir,
                );
                STDOUT->print( qq{Recovered $release in /efs/dist in $cell\n} );
            } else {
                STDOUT->print( qq{Nothing to recover for $release in /efs/dist in $cell\n} );
            }

        }

    }

    return 1;
}

sub purge {

    my $class = shift;

    my %args = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type     => SCALAR | OBJECT },
        project  => { type     => SCALAR | OBJECT },
        name     => { type     => SCALAR | OBJECT },
        now      => { type     => BOOLEAN,
                      optional => 1 },
        %EFS::CLI::validate_purge_args,
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        destroyed => 1,
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{purge_rpc},
        args   => [ %args ],
    );

}

sub purge_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{name},
        destroyed => 1,
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project_dir = $release->project->get_path(q{dev});

    if ( not $release->destroyed ) {
        die qq{Release '$release' has not been destroyed.\n};
    }

    my $destroyed = $release->destroyed;

    if ( not $args{authorized} and $args{now} and
             $release->get_diststate->cells_partially_updated ) {
        die(
            qq{Only efsops can perform purges with -now once the\n},
            qq{release has been disted anywhere at all.\n},
        );
    }

    if ( not $args{now} ) {

        Readonly my $NUM_DAYS => 14;

        if ( $release->delta_days( $destroyed ) < $NUM_DAYS ) {
            STDOUT->print(
                qq{Unable to purge '$release'.\n},
                qq{It was destroyed less than $NUM_DAYS days ago ($destroyed).\n}
            );
            return 1;
        }

    }

    my $dirlock = EFS::DirLock->new({
        directory => $project_dir,
        retry     => 1,
        command   => $history->request,
    });

    my $diststate = $release->get_diststate;

    my @cells = grep { not $_->hidden } $diststate->cells_partially_updated;

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell         => $cell,
                class        => $class,
                method       => q{purge_cell},
                args         => [
                    cell     => $cell->name,
                    metaproj => $release->metaproj->name,
                    project  => $release->project->name,
                    release  => $release->name,
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to purge '$release' from the following cells:\n\t},
        join( qq{\n\t}, $fm_cells->get_failed ) . qq{\n}
    );

    if ( $release->uploaded ) {
        my $package = EFS::Package->new(
            history => $args{history},
            metaproj  => $args{metaproj},
            project   => $args{project},
            name      => $args{name},
        );
        $package->purge;
    }

    given ( $release->cell->fstype ) {
        when ( q{afs} ) {
            die qq{Unable to purge releases in AFS cells yet\n};
            # XXX: remove the volumes, release AND installs
        }
        when ( q{nfs} ) {
            my $dev_path          = $release->get_path(q{dev});
            my $destroyed_dir     = $release->project->get_path(q{dev})  . q{/.destroyed};
            my $destroyed_release = $destroyed_dir .q{/}. basename($dev_path);

            if ( -d $destroyed_release ) {
                EFS::Util->RemoveDirHierarchy( top => $destroyed_release );
                STDOUT->print( qq{Purged '$release' from /efs/dev\n} );
            }
        }
    }

    foreach my $install ( $release->installs ) {
        $install->auditstates->delete_all;
        $install->cellstates->delete_all;
        $install->delete;
    }

    foreach my $dependency ( $release->dependencies ) {
        $dependency->delete;
    }

    my $object_id = $release->object_id;

    $release->attrs->delete_all;
    $release->delete;

    STDOUT->print( qq{Successfully purged '$object_id' from all cells\n} );

    return 1;

}

sub purge_cell {

    my $class = shift;

    my (%args) = validate( @_, {
        cell     => { type => SCALAR },
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        release  => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $resultset = $class->get_resultset;
    my $release = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        project   => $args{project},
        name      => $args{release},
        destroyed => 1,
    );

    my $project = $release->project;

    given ( $cell->fstype ) {

        when ( q{afs} ) {
            die qq{Unable to purge releases in AFS cells yet\n};
        }

        when ( q{nfs} ) {
            my $admin_path = $cell->admin_path( type => q{dist} );
            my @destroy = (
                qq{$admin_path/dist/$project/.destroyed/} . $release->name,
                glob(qq{$admin_path/*/dist/$release}),
            );
            if ( not @destroy ) {
                STDOUT->print(
                    qq{Release $release was never disted to $cell.\n},
                    qq{Nothing to purge.\n},
                );
                return 1;
            }
            EFS::Util->RemoveDirHierarchies( tops => \@destroy );
        }

    }

    STDOUT->print( qq{Purged '$release' from /efs/dist in $cell\n} );

    return 1;

}

sub setstage {

    my $class = shift;

    my %args = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type               => SCALAR },
        project  => { type               => SCALAR },
        name     => { type               => SCALAR },
        stage    => { type               => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{setstage_rpc},
        args   => [ %args ],
    );

}

sub setstage_rpc {

    my $class = shift;
    my %args  = @_;

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $release_stage = $release->stage;
    my $release_dir   = $release->get_path(q{dev});

    my $dirlock = EFS::DirLock->new({
        directory => $release_dir,
        retry     => 1,
        command   => $history->request,
    });

    if ( $release_stage eq $args{stage} ) {
        STDOUT->print( qq{Release stage is already set to $args{stage}\n} );
        return 1;
    }

    given ( qq{$release_stage/$args{stage}} ) {

        when ( [qw( deprecated/dev deprecated/qa deprecated/rejected
                    dev/deprecated qa/deprecated rejected/deprecated
                    rejected/prod  prod/rejected )] ) {
            die qq{Unsupported stage transition $release_stage -> $args{stage}\n};
        }

        when ( [qw( deprecated/prod prod/deprecated
                    rejected/dev    dev/rejected
                    rejected/qa     qa/rejected
                    dev/qa          qa/dev )] ) {
            break;
        }

        when ( [qw( dev/prod qa/prod )] ) {

            $release->validate_checkpoint;
            $release->validate_audits;
            $release->validate_prod_dependencies;

            if ( not $release->locked ) {

                my $feature_rs = $class->get_resultset( q{Feature} );

                if ( $release->name eq q{incr} ) {

                    if ( $feature_rs->configured(q{allow_insecure_prod_incr}) ) {
                        warn(
                            qq{Setting the unlocked release $release stage to prod,\n},
                            qq{since the allow_insecure_prod_incr feature is enabled.\n},
                        );
                    } else {
                        die(
                            qq{Unable to set the stage to prod, because the release is not locked.\n}
                        );
                    }

                } else {

                    my $feature = $feature_rs->configured( q{allow_insecure_disable_lockdown} );

                    if ( $feature && $release->find_mpr_attr(q{disable_lockdown}) ) {
                        warn(
                            qq{Setting the unlocked release $release stage to prod,\n},
                            qq{since the allow_insecure_disable_lockdown feature is enabled.\n},
                        );
                    } else {
                        die(
                            qq{Unable to set the stage to prod, because the release is not locked.\n}
                        );
                    }

                }

            }

            $release->get_diststate->cells_partially_updated || die(
                qq{Unable to set the stage to prod, because the release has not been disted yet.\n}
            );

        }

        when ( [qw( prod/dev prod/qa )] ) {

            # [1] No dependencies can be stage=prod/deprecated

            my @dependents = $release->all_dependents;

            if ( grep { $_->dependent_release->has_prod_stages } @dependents ) {
                die(
                    qq{Release '$release' has prod/deprecated dependents,\n},
                    qq{and therefore can NOT be reverted to dev or qa.\n}
                );
            }

            # [2] Never been disted to ANY prod cells (need the
            # Diststate objects to implement this)

            my $diststate = $release->get_diststate;

            if ( grep { $_->celltype eq q{prod} } $diststate->cells_partially_updated ) {
                die(
                    qq{Release '$release' has been disted to one or more\n},
                    qq{prod cells and can NOT be reverted to dev or qa.\n},
                );
            }

        }

        default {
            die qq{Unrecognized stage transition $release_stage -> $args{stage}\n};
        }

    }

    $release->set_column( stage => $args{stage} );
    $release->update;
    STDOUT->print(
        qq{Changed the stage of $release from $release_stage to $args{stage}\n}
    );

    if ( my $clone = $release->cloneof ) {
        if ( $release->has_prod_stages and not $clone->has_prod_stages ) {
            warn(
                qq{Release $release is a clone of $clone, but\n},
                qq{$clone has a stage of '}, $clone->stage, qq{'\n},
                qq{Cloning optimizations will not be available when\n},
                qq{$release is disted to prod cells\n},
            );
        }
    }

    return 1;

}

sub audit {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
        audits   => { type => SCALAR, default  => q{required}, optional => 1 },
        force    => { type => BOOLEAN, optional => 1 },
        verbose  => { type => BOOLEAN, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $release->cell,
        class  => $class,
        method => q{audit_rpc},
        args   => [ %args ],
    );

}

sub audit_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $release_dir = $release->get_path(q{dev});

    my @installs = sort $release->installs;

    if ( not @installs ) {
        STDOUT->print(
            qq{Unable to audit release '$release', since it has no installs.\n}
        );
        return 1;
    }

    my $dirlock = EFS::DirLock->new({
        directory => $release_dir,
        retry     => 1,
        command   => $history->request,
    });

    STDOUT->print(
        qq{Performing $args{audits} audits for all installs },
        qq{in release '$release'\n}
    );

    my $fm_installs = EFS::ForkManager->new({

        list => \@installs,

        action => sub {

            my $install = shift;

            EFS::CLI::Install->audit(
                history         => $args{history},
                metaproj        => $args{metaproj},
                project         => $args{project},
                release         => $args{name},
                name            => $install->name,
                audits          => $args{audits},
                force           => $args{force},
                verbose         => $args{verbose},
                _release_locked => 1,
            );

        },

    });

    $fm_installs->run || die(
        qq{Unable to successfully audit the following installs:\n\t},
        join( qq{\n\t}, $fm_installs->get_failed ), qq{\n}
    );

    STDOUT->print( qq{Successfully audited all installs in release '$release'\n} );

    return 1;

}

sub set_distgroup {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR },
        project   => { type => SCALAR },
        name      => { type => SCALAR },
        distgroup => { type => SCALAR },
    });

    my $feature_rs = $class->get_resultset( q{Feature} );

    if ( $feature_rs->configured( q{disable_restricted_access} ) ) {
        die(
            qq{Unable to set the distgroup attribute, since the\n},
            qq{disable_restricted_access feature has been enabled.\n},
        );
    }

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $distgroup = EFS::Groupname->new({ name => $args{distgroup} });

    if ( $release->distgroup eq $distgroup->name ) {
        STDOUT->print( qq{Release $release already has distgroup $distgroup\n} );
    } else {
        $release->set_update( distgroup => $distgroup->name );
        $release->update;
        STDOUT->print( qq{Set distgroup for release $release to $distgroup\n} );
    }

    return 1;

}

sub clear_distgroup {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type       => SCALAR },
        project  => { type       => SCALAR },
        name     => { type       => SCALAR },
    });

    my $feature_rs = $class->get_resultset( q{Feature} );

    if ( $feature_rs->configured( q{disable_restricted_access} ) ) {
        die(
            qq{Unable to set the distgroup attribute, since the\n},
            qq{disable_restricted_access feature has been enabled.\n},
        );
    }

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    if ( $release->distgroup ) {
        $release->set_column( distgroup => undef );
        $release->update;
        STDOUT->print( qq{Cleared distgroup for release $release\n} );
    } else {
        STDOUT->print( qq{Release $release has no distgroup\n} );
    }

    return 1;

}

sub set_noworldaccess {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj      => { type => SCALAR },
        project       => { type => SCALAR },
        name          => { type => SCALAR },
        noworldaccess => { type => SCALAR },
    });

    my $feature_rs = $class->get_resultset( q{Feature} );

    if ( $feature_rs->configured( q{disable_restricted_access} ) ) {
        die(
            qq{Unable to set the noworldaccess attribute, since the\n},
            qq{disable_restricted_access feature has been enabled.\n},
        );
    }

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $noworldaccess = $args{noworldaccess} ? 1 : 0;

    $release->set_column( noworldaccess => $noworldaccess );
    $release->update;

    STDOUT->print( qq{Set noworldaccess for release $release to $noworldaccess\n} );

    return 1;

}

sub clone {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
        clone    => { type => SCALAR },
        src      => { type => SCALAR, optional => 1 },
        build    => { type => SCALAR, optional => 1 },
        force    => { type => SCALAR, optional => 1 },
    });

    my $project_rs = $class->get_resultset( q{Project} );
    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{clone_rpc},
        args   => [ %args ],
    );

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return $release;

}

sub clone_rpc {

    my $class = shift;

    my %args = @_;

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project_rs = $class->get_resultset( q{Project} );
    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    my $project_lock = EFS::DirLock->new({
        directory => $project->get_path(q{dev}),
        retry     => 1,
        command   => $history->request,
    });

    my $resultset = $class->get_resultset;
    my $clone = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{clone},
    );

    if ( not $clone->locked ) {
        die qq{Clone $clone is not locked\n};
    }

    if ( length( $args{clone} ) ne length( $args{name} ) ) {
        die(
            qq{Cloned releases must have names of matching length\n},
            qq{length($args{clone}) != length($args{name})\n},
        );
    }

    my $clone_lock = EFS::DirLock->new({
        directory => $clone->get_path(q{dev}),
        retry     => 1,
        command   => $history->request,
    });

    my $release = $resultset->retrieve_bynames(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    if ( $release ) {

        if ( $release->locked ) {
            die qq{Release $release is locked\n};
        }

        if ( not $args{force} ) {
            die qq{Release $release exists.  Use --force to re-clone.\n};
        }

    }
    else {

        $release = EFS::CLI::Release->create_rpc(
            history  => $args{history},
            metaproj => $args{metaproj},
            project  => $args{project},
            name     => $args{name},
            _project_locked => 1,
        );

    }

    my $release_lock = EFS::DirLock->new({
        directory => $release->get_path(q{dev}),
        retry     => 1,
        command   => $history->request,
    });

    my $clone_dev    = $clone->get_path(q{dev});
    my $release_dev  = $release->get_path(q{dev});
    my $clone_dist   = $clone->get_path(q{dist});
    my $release_dist = $release->get_path(q{dist});

    foreach my $clone_install ( $clone->installs ) {

        my $install = EFS::CLI::Install->create_rpc(
            history         => $args{history},
            metaproj        => $args{metaproj},
            project         => $args{project},
            release         => $args{name},
            name            => $clone_install->name,
            _release_locked => 1,
        );

        EFS::Util->CloneDirectory(
            from    => $clone_install->get_path(q{dev}),
            to      => $install->get_path(q{dev}),
            replace => {
                $clone_dist => $release_dist,
                $clone_dev  => $release_dev,
            },
        );

        EFS::CLI::Install->_lock_unlock_rpc(
            history         => $args{history},
            metaproj        => $args{metaproj},
            project         => $args{project},
            release         => $args{name},
            name            => $clone_install->name,
            lockmode        => q{unlock},
            _release_locked => 1,
        );

        $install->set_column( cloneof => $clone_install->id );
        $install->update;

        STDOUT->print( qq{Cloned install $install from $clone_install\n} );

    }

    if ( $args{src} ) {
        EFS::Util->CloneDirectory(
            from    => qq{$clone_dev/src},
            to      => qq{$release_dev/src},
            replace => {
                $clone_dist => $release_dist,
                $clone_dev  => $release_dev,
            },
        );
    }

    if ( $args{build} ) {
        EFS::Util->CloneDirectory(
            from    => qq{$clone_dev/build},
            to      => qq{$release_dev/build},
            replace => {
                $clone_dist => $release_dist,
                $clone_dev  => $release_dev,
            },
        );
    }

    if ( $args{src} or $args{build} ) {
        EFS::CLI::Release->_lock_unlock_rpc(
            history         => $args{history},
            metaproj        => $args{metaproj},
            project         => $args{project},
            name            => $args{name},
            lockmode        => q{unlock},
            norecurse       => 1,
            _release_locked => 1,
        );
    }

    $release->set_column( cloneof => $clone->id );
    $release->update;

    my $clone_diststate = $clone->get_diststate;

    my @missing =
        grep { not $_->hidden } $clone_diststate->cells_needing_update;

    if ( @missing ) {
        warn(
            qq{Clone $clone has not been updated in the following cells:\n\t},
            join(qq{\n\t}, map { $_->name } @missing ), qq{\n},
            qq{Clone dist optimizations will not work in those cells,\n},
            qq{unless the clone release is disted first.\n},
        );
    }

    STDOUT->print(qq{Cloned release $clone to $release\n});

    return 1;

}

sub unclone {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $release = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    $release->unclone;

    STDOUT->print( qq{Release $release is no longer a clone\n} );

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
