package EFS::DBI::Dependency;

use strict;
use warnings;
use English;

use Params::Validate qw( :all );
use Readonly;

Readonly my $SCREEN_WIDTH => 1024;

# metaproj (32) + project (64) + release(alias) (128) + pair of slashes (2)
Readonly my $RELEASE_OBJECT_ID_MAXLENGTH => 226;
Readonly my $TYPE_MAXLENGTH => 32;

sub sort_id {
    my $self = shift;
    return join( qq{\n},
                 sprintf( q{%-*s}, $TYPE_MAXLENGTH,
                          $self->type ),
                 sprintf( q{%-*s}, $RELEASE_OBJECT_ID_MAXLENGTH,
                          $self->dependent_release->object_id ),
                 sprintf( q{%-*s}, $RELEASE_OBJECT_ID_MAXLENGTH,
                          $self->target->object_id ) ) . qq{\n};
}

sub stage {
    return shift->target->stage;
}

sub dependent_object_id {
    return shift->dependent_release->object_id;
}

sub dependency_object_id {
    return shift->target->object_id;
}

sub get_diststate {
    return shift->target->get_diststate;
}

sub check_circular {

    my $class = shift;

    my (%args) = validate( @_, {
        dependencies => { type => HASHREF },
        dependents   => { type => HASHREF },
    });

    my %dependencies = %{ $args{dependencies} };
    my %dependents   = %{ $args{dependents} };

    my $pass = 1;

  DEPENDENCIES: while ( 1 ) {

        my $removed = 0;

        foreach my $dependent_mpr ( sort keys %dependencies ) {

            next if exists $dependents{ $dependent_mpr };

            foreach my $dependency_mpr ( keys %{ $dependencies{ $dependent_mpr } } ) {

                delete $dependents{ $dependency_mpr }->{ $dependent_mpr };

                if ( not keys %{ $dependents{ $dependency_mpr } } ) {
                    delete $dependents{ $dependency_mpr };
                }

            }

            delete $dependencies{ $dependent_mpr };

            $removed++;

        }

        if ( $removed ) {
            $pass++;
        } else {
            last DEPENDENCIES;
        }

    }

    if ( not %dependencies ) {
        return 1;
    }

    # Now we do it again, but from the bottom up.

    $pass       = 1;

  DEPENDENTS: while ( 1 ) {

        my $removed = 0;

        foreach my $dependency_mpr ( sort keys %dependents ) {

            next if exists $dependencies{ $dependency_mpr };

            foreach my $dependent_mpr ( keys %{ $dependents{ $dependency_mpr } } ) {

                delete $dependencies{ $dependent_mpr }->{ $dependency_mpr };

                if ( not keys %{ $dependencies{ $dependent_mpr } } ) {
                    delete $dependencies{ $dependent_mpr };
                }

            }

            delete $dependents{ $dependency_mpr };

            $removed++;

        }

        if ( $removed ) {
            $pass++;
        } else {
            last DEPENDENTS;
        }

    }

    # Whatever is left in the %dependencies and %dependents hashes
    # could not be removed because it must be part of a circular
    # dependency!!

    # XXX: Once we have the tree display working, we can make this
    # look better.

    die(
        qq{Found one or more circular dependencies in the following data:\n},
        Data::Dumper->Dump([\%dependencies],[q{dependencies}]),
    );

    return;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
