package EFS::Diststate;

use strict;
use warnings;
use English;

use Moose;
use Params::Validate qw( :all );
use List::Util qw( min max );
use List::MoreUtils qw( uniq all );
use Set::Scalar;
use EFS::Colors;
use Lingua::EN::Inflect qw( PL );
use Readonly;

use EFS::Diststate::Any;
use EFS::Diststate::Metaproj;
use EFS::Diststate::Project;
use EFS::Diststate::Release;
use EFS::Diststate::ReleaseAlias;
use EFS::Diststate::Install;

my $EMPTY_STRING = q{};
Readonly my $SCREEN_WIDTH => 1024;


sub BUILD {

    my $self = shift;

    # This is an abstract class!
    if ( ref($self) eq __PACKAGE__ ) {
        die( q{Attempt to instantiate abstract class}
                 . __PACKAGE__
                     . qq{directly\n} );
    }

    return;

}

sub name {
    return shift->related_item->name;
}

sub cells_needing_update {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return uniq sort map { $_->cells_needing_update(%args) }
        $self->subitem_diststates;
}

# Returns "install", "metaproj", "release", etc.
sub related_item_type {
    return shift->related_item->class2type;
}

sub installs_needing_update {
    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return
        map { $_->installs_needing_update(%args) }
            $self->subitem_diststates;
}

sub releases_needing_update {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return
        map { $_->releases_needing_update(%args) }
            $self->subitem_diststates;

}

sub projects_needing_update {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return
        map { $_->projects_needing_update(%args) }
            $self->subitem_diststates;

}

sub metaprojects_needing_update {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return
        map { $_->metaprojects_needing_update(%args) }
            $self->subitem_diststates;

}

sub needs_update {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    # If any sub-diststate needs update, then we immediately
    # know that $self does too
    foreach my $diststate ( $self->subitem_diststates ) {
        next if $diststate->is_irrelevant;
        if ( $diststate->needs_update(%args) ) {
            return 1;
        }
    }

    return;

}

sub has_been_updated {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    my @subitem_diststates = $self->subitem_diststates;
    if ( not @subitem_diststates ) {
        return;
    }

    # If any sub-diststate is not updated, then we immediately
    # know $self's diststate is not updated either
    foreach my $diststate (@subitem_diststates) {
        next if $diststate->is_irrelevant;
        if ( not $diststate->has_been_updated(%args) ) {
            return;
        }
    }

    return 1;

}

sub cells_updated {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    # Only cells common to *all* sub-diststates can be
    # considered updated for $self's diststate.
    # So we create a set for the cells in the first sub-diststate,
    # then find the common intersection with the others.

    my @sub_diststate_cells =
        map { [ $_->cells_updated(%args) ] }
            $self->subitem_diststates;

    if ( not @sub_diststate_cells ) {
        return;
    }

    my $cell_set = Set::Scalar->new( @{ shift @sub_diststate_cells } );

    for my $cell_arrayref (@sub_diststate_cells) {
        $cell_set = $cell_set->intersection(
            Set::Scalar->new( @{ $cell_arrayref } ) );
    }

    return $cell_set->members;

}

# "at least partially updated" might be more accurate--this
# returns the union, rather than the intersection as above
sub cells_partially_updated {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return uniq sort map { $_->cells_updated(%args) }
        $self->subitem_diststates;

}

sub cells_never_updated {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
        },
    });

    my %partials =
        map { $_->name => 1 } $self->cells_partially_updated(%args);

    return @{[ sort grep { not $partials{$_->name} } @{ $args{cells} } ]};

}

sub installs_updated {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return
        map { $_->installs_updated(%args) }
            $self->subitem_diststates;

}

sub releases_updated {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return
        map { $_->releases_updated(%args) }
            $self->subitem_diststates;

}

sub projects_updated {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return
        map { $_->projects_updated(%args) }
            $self->subitem_diststates;

}

sub metaprojects_updated {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1
        },
    });

    return
        map { $_->metaprojects_updated(%args) }
            $self->subitem_diststates;

}

sub is_destroyed { return shift->related_item->is_destroyed; }

# Some diststates aren't relevant to their enclosing item's
# diststate.  For example, destroyed things and releases with
# no installs.
sub is_irrelevant {

    my $self = shift;

    # Destroyed things are irrelevant to the larger item's diststate.
    if ($self->is_destroyed) {
        return 1;
    }

    # Items with no installs are irrelevant too.
    if ($self->related_item->can(q{installs})) {
        return not $self->related_item->installs->count;
    }
    else {
        # Items that can't have installs (Installs or ReleaseAlias)
        # must be relevant if not destroyed
        return;
    }

}

sub display_text_table {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type => ARRAYREF,
        },
        verbose => {
            type    => BOOLEAN | UNDEF,
            default => 0,
        },
        respect_cell_ordering => {
            type    => BOOLEAN | UNDEF,
            default => 0,
        },
    });

    # We're going to modify the cell list, so make it a copy
    $args{cells} = [ @{$args{cells}} ];

    my $subitems_by_type = $self->subitem_diststates_by_type;

    my $legend_needed = 0;

    my %cells_by_type;
    foreach my $cell ( @{ $args{cells} } ) {
        push @{ $cells_by_type{$cell->celltype} }, $cell;
    }

    my $cell_rs = EFS::DBI->get_resultset( q{Cell} );

  CELLTYPE:
    foreach my $cell_type ( sort keys %cells_by_type ) {

        my $output_needed = $args{verbose};
        my @output;

        my @cells = @{ $cells_by_type{$cell_type} };

        if ( not $args{respect_cell_ordering} ) {
            @cells = EFS::DBI->sort_objects(@cells);
        }

        my @table_format = map { q{l} } (q{item}, @cells);

        push @output, ( $self->_celltype_marker($cell_type) );

      ITEMTYPE:
        foreach my $type ( sort keys %{ $subitems_by_type } ) {

            my @subitems = @{ $subitems_by_type->{$type} };

            my $table = Text::FormatTable->new(
                join( q{  }, @table_format )
            );

            my @headers = (
                ucfirst( $type ),
                map { color_neutral($_->shortname) } @cells
            );

            my $table_needed = $args{verbose};
            $table->head(@headers);
            $table->rule(q{=});

            my $subitems_up_to_date = 0;
            my $subitems_without_cp = 0;
            my @subitems_destroyed;

          SUBITEM:
            for my $subitem ( sort { lc($a->name) cmp lc($b->name) }
                                  @subitems ) {

                my @statuses
                    = map { $subitem->_text_status( cells => [$_] ) }
                        @cells;
                my $disted_text = _disted_text();
                my $nocp_text = _nocp_text();
                my $wrong_stage_text = _wrong_stage_text();

                if ($subitem->is_destroyed) {
                    push @subitems_destroyed, $subitem;
                    next SUBITEM;
                }
                if (all { $_ eq $wrong_stage_text } @statuses) {
                    next SUBITEM unless $args{verbose};
                }

                if (all { $_ eq $disted_text } @statuses) {
                    $subitems_up_to_date++;
                    next SUBITEM unless $args{verbose};
                }

                if ( all { $_ eq $nocp_text } @statuses ) {
                    $subitems_without_cp++;
                    next SUBITEM unless $args{verbose};
                }

                $table->row( $subitem->name, @statuses );
                $table_needed++;

            }

            if ($table_needed) {
                $legend_needed++;

                $table->rule;
                push @output, ( $table->render($SCREEN_WIDTH) );
                $output_needed++;
            }

            if ($output_needed) {

                if ( $args{verbose} ) {
                    if (@subitems_destroyed) {
                        # we need to print destroyed items
                        push @output, (
                            _destroyed_count_text(
                                scalar @subitems_destroyed, $type)
                                . qq{:\n}
                            );

                        @subitems_destroyed =
                            sort { lc($a->name) cmp lc($b->name) }
                                @subitems_destroyed;

                        for my $item ( map { $_->name }
                                           @subitems_destroyed ) {
                            push @output, ( qq{$item\n} );
                        }

                    }
                } else { # we print a count of items we didn't print
                    push @output, $self->_table_counts({
                        item                => $type,
                        up_to_date_count    => $subitems_up_to_date,
                        no_checktime_count  => $subitems_without_cp,
                        destroyed_count     =>
                            scalar @subitems_destroyed,
                    });
                }
            }
        }

        if ($output_needed) {
            STDOUT->print(@output);
        }
        else {
            STDOUT->print(
                $self->_celltype_uptodate_message($cell_type));
            # Remove these cells from %args so they won't be
            # displayed in the cell equivalence legend
            @{$args{cells}} =
                grep { $_->celltype ne $cell_type } @{$args{cells}};
        }
    }

    if ( $legend_needed ) {
        STDOUT->print(
            $self->_table_legend( verbose => $args{verbose} )
        );
    }

    return 1;

}

# This is used by ReleaseAlias::Diststate too, so it's factored out
sub _celltype_marker {
    my $self      = shift;
    my $cell_type = shift;
    return (ucfirst($cell_type) . qq{ environment diststate:\n});
}

# Ditto
sub _celltype_uptodate_message {
    my $self      = shift;
    my $cell_type = shift;
    return (ucfirst($cell_type) . qq{ environment is up to date.\n});
}

# This should only come into play with project diststates, which have
# both releases and releasealiases as subitem diststates, but it will
# future-proof us against any other cases like this
sub subitem_diststates_by_type {
    my $self     = shift;
    my @subdists = $self->subitem_diststates;

    my $result;
    for my $subdist (@subdists) {
        my $type = $subdist->related_item_type;
        push @{ $result->{$type} }, $subdist;
    }

    # Now we have a structure like this:
    # {
    #    release => [ $distate1, $diststate2, ... ],
    #    releasealias => [ $distate1, $diststate2, ... ]
    # }
    return $result;
}

# The checktime of anything higher-level than an install
# is the latest checktime seen
# XXX: This is probably not pedantically correct, but it works
#      for the current usecase.  Better perhaps to create a
#      new method to ask, "are there any checktimes below here?"
sub get_checktime {
    my $self = shift;
    return max map { $_->get_checktime || 0 } $self->subitem_diststates;
}

sub _disted_text        { return color_good(q{D}) }
sub _nocp_text          { return color_neutral(q{-}) }
sub _needs_update_text  { return color_bad(q{X}) }
sub _needs_partial_text { return color_warn(q{P}) }
sub _wrong_stage_text   { return color_warn(q{S}) }

sub _text_status {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => {
            type     => ARRAYREF,
            optional => 1,
        },
    });

    my $updated      = $self->has_been_updated(%args);
    my $needs_update = $self->needs_update(%args);

    if ( not $self->get_checktime ) {
        # nothing there
        return _nocp_text();
    }
    elsif ( $updated and $needs_update ) {
        return _needs_partial_text();
    }
    elsif ($needs_update) {
        return _needs_update_text();
    }
    else {
        if (not $self->related_item->has_prod_stages
                and $args{cells}->[0]->celltype eq q{prod}) {
            return _wrong_stage_text();
        }
        # disted
        return _disted_text();
    }

    return die qq{Weird, no dist status?\n};

}

sub _table_counts {

    my $self = shift;

    my (%args) = validate( @_, {
        up_to_date_count => {
            type    => SCALAR,
            default => 0,
        },
        no_checktime_count => {
            type    => SCALAR,
            default => 0,
        },
        destroyed_count => {
            type    => SCALAR,
            default => 0,
        },
        item => {
            type    => SCALAR,
            regex   => qr/\A \w+ \z/xms, # a single word only
            default => q{item},
        },
    });

    my $item = $args{item};

    my $output = $EMPTY_STRING;
    if (my $n = $args{up_to_date_count}) {
        $output .= $n . PL(qq{ $item}, $n) . PL(q{ is }, $n) .
            qq{up to date and not displayed.\n};
    }
    if (my $n = $args{no_checktime_count}) {
        $output .= $n . PL(qq{ $item}, $n) . PL(q{ has }, $n) .
            q{no checktime (may be empty) and } .
                qq{not displayed.\n};
    }
    if (my $n = $args{destroyed_count}) {
        $output .= _destroyed_count_text($n, $item)
            . qq{ and not displayed.\n};
    }
    return $output;

}

sub _destroyed_count_text {
    my ($n, $item) = @_;
    if ($n) {
        return $n . PL(qq{ $item}, $n) . PL(q{ has }, $n)
            . q{been destroyed but not yet purged};
    }
    return $EMPTY_STRING;
}

sub _table_legend {

    my $self = shift;

    my (%args) = validate( @_, {
        verbose => {
            type => BOOLEAN | UNDEF,
            default => 0
        },
    });

    my %description_for = (
        _disted_text()        => q{Item is disted and up to date},
        _needs_partial_text() => q{Item is disted but needs update},
        _needs_update_text()  => q{Item has not been disted},
    );

    if ( $args{verbose} ) {
        $description_for{_nocp_text()} =
            q{Item has no checktimes (may be empty)};
        $description_for{_wrong_stage_text()} =
            q{Item is in wrong stage for cell};
    }

    my $output = qq{Legend:\n};

    for my $text (sort keys %description_for) {
        $output .= qq{  $text: } . $description_for{$text} . qq{\n};
    }

    return $output;

}


1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
