package EFS::Request;

use strict;
use warnings;
use English;
use feature q{switch};

use EFS::FQDN qw( fqdn );

use Moose;
use Params::Validate qw( :all );
use Storable qw(thaw);
use Socket;
use Authen::PAM;
use POSIX;
use Try::Tiny;

use MIME::Base64;

has q{client}   => ( is => q{rw}, isa => q{IO::Socket::SSL}, required => 1 );
has q{conntype} => ( is => q{rw}, isa => q{Str} );

has q{authhost}     => ( is => q{rw}, isa => q{Str} );
has q{authuser}     => ( is => q{rw}, isa => q{Str} );
has q{authpass}     => ( is => q{rw}, isa => q{Str} );
has q{authtype}     => ( is => q{rw}, isa => q{Str} );
has q{authrequired} => ( is => q{rw}, isa => q{Int}, default => 1 );

has q{metadata}  => ( is => q{rw}, isa => q{HashRef}, default  => sub { {} } );
has q{operation} => ( is => q{rw}, isa => q{HashRef}, default  => sub { {} } );
has q{logfile}   => ( is => q{rw}, isa => q{EFS::Logfile} );

has q{history_args} => ( is => q{rw}, isa => q{HashRef} );
has q{history_obj}  => ( is => q{rw}, isa => q{EFS::DBI::Result::History} );

sub BUILD {

    my $self = shift;

    $self->authenticate;

    my $subclass = q{EFS::Request::} . $self->conntype;

    bless $self, $subclass;

    $self->parse_request_data;
    $self->create_logfile;
    $self->clone_history;

    return;

}

sub authenticate {

    my $self  = shift;

    my $client = $self->client;

    $self->validate_client;

    my $conntype = $client->getline or $self->fatal( qq{No connection type provided\n} );
    chomp $conntype;
    $self->conntype( $conntype );

    my $authtype = $client->getline or $self->fatal( qq{No authentication type provided\n} );
    chomp $authtype;
    $self->authtype( $authtype );

    EFS::Syslog->message( q{info}, qq{Accepted $authtype/$conntype connection} );

    my $config = EFS::Config::Server->new;

    given ( qq{$authtype/$conntype} ) {

        when ( q{GSSAPI/RPC} ) {
            $client->peer_certificate or $self->fatal( qq{No peer certificate\n} );
            $self->authenticate_gssapi_server;
        }

        when ( q{SSL/RPC} ) {
            $client->peer_certificate or $self->fatal( qq{No peer certificate\n} );
            $self->authuser( q{efsd} );
        }

        when ( q{GSSAPI/CLI} ) {
            if ( $config->get_value( q{security}, q{authtype} ) !~ m{\b$authtype\b}msx ) {
                $self->fatal( qq{This server does not support authtype '$authtype'\n} );
            }
            $self->authenticate_gssapi_server;
        }

        when ( q{PAM/CLI} ) {
            my $authuser = $client->getline or $self->fatal( qq{No username provided\n} );
            my $authpass = $client->getline or $self->fatal( qq{No password provided\n} );
            if ( $config->get_value( q{security}, q{authtype} ) !~ m{\b$authtype\b}msx ) {
                $self->fatal( qq{This server does not support authtype '$authtype'\n} );
            }
            chomp( $authuser, $authpass );
            $self->authenticate_pam_server( $authuser, $authpass );
        }

        default {
            $self->fatal(qq{Invalid EFS protocol: '$authtype/$conntype'\n});
        }

    }

    return;

}

sub validate_client {

    my $self  = shift;

    my $client = $self->client;

    my $client_sockaddr = $client->peeraddr ||
        $self->fatal( qq{Unable to determine peeraddr of client socket.\n} );

    my $client_address = $client->peerhost ||
        $self->fatal( qq{Unable to determine peerhost of client socket.\n} );

    if ( $client_address eq q{127.0.0.1} ) {
        $self->authhost( fqdn );
        return 1;
    }

    my ($client_hostname) =
        gethostbyaddr( $client_sockaddr, AF_INET ) or
            $self->fatal( qq{Unable to determine hostname of $client_address: $ERRNO\n} );

    my @gethostbyname =
        gethostbyname($client_hostname) or
            $self->fatal( qq{Unable to determine addresses of $client_hostname: $ERRNO\n} );

    my $found_match = 0;

    my @addrs = map { inet_ntoa($_) } @gethostbyname[ 4 .. $#gethostbyname ];

    foreach my $addr (@addrs) {
        if ( $addr eq $client_address ) {
            $found_match++;
            last;
        }
    }

    if ( not $found_match ) {
        $self->fatal(
            qq{Invalid client address $client_address\n},
            qq{'$client_address' resolves to $client_hostname,\n},
            qq{but that resolves to the following addresses:\n\t},
            join ( qq{\n\t}, @addrs ),
            qq{\n},
        );
    }

    if ( $client->peer_certificate ) {
        $client->verify_hostname( $client_hostname, q{smtp} ) or
            $self->fatal( qq{Unable to verify hostname of client certificate\n} );
    }

    $self->authhost($client_hostname);

    return 1;

}

sub parse_request_data {

    my $self   = shift;
    my $client = $self->client;

    my $buffer;

    if ( read ( $client, $buffer, 4 ) != 4 ) {
        $self->fatal(qq{Unable to read size of incoming message: $ERRNO\n});
    }

    my $want = unpack ( q{N}, $buffer );

    my $got = read ( $client, $buffer, $want );

    if ( $got != $want ) {
        $self->fatal(qq{Unable to read incoming message: $ERRNO\n});
    }

    my $request;

    try { $request = thaw($buffer); }
    catch {
        $self->fatal(qq{Unable to thaw frozen request data: $_\n});
    };

    if (   ( not ref $request eq q{HASH} )
               or ( scalar keys %{$request} != 2 ) ) {
        $self->fatal(qq{Invalid request: not a HASH ref with two keys\n});
    }

    foreach my $section ( qw( metadata operation ) ) {
        if ( not ref $request->{$section} eq q{HASH} ) {
            $self->fatal(qq{Invalid request: missing $section section\n});
        }
    }

    my $metadata  = $request->{metadata};
    my $operation = $request->{operation};

    foreach my $required ( $self->metadata_required ) {
        if ( not $metadata->{$required} ) {
            $self->fatal(
                q{Invalid request: missing required },
                qq{metadata key '$required'\n}
            );
        }
    }

    foreach my $required ( $self->operation_required ) {
        if ( not $operation->{$required} ) {
            $self->fatal(
                q{Invalid request: missing required },
                qq{operation key '$required'\n}
            );
        }
    }

    if ( $self->isa(q{EFS::Request::CLI}) ) {

        my $object_type = $operation->{object_type};

        $operation->{object_class}
            = q{EFS::CLI::} . EFS::DBI->type2subclass($object_type);

        if ( not $EFS::CLI::commands{$object_type} ) {
            $self->fatal(qq{Unsupported EFS object type '$object_type'\n});
        }

        my $action = $operation->{action};

        if ( not $EFS::CLI::commands{$object_type}->{$action} ) {
            $self->fatal(
                qq{Unsupported action '$action' for },
                qq{EFS object type '$object_type'\n}
            );
        }


        if ( grep { $_ eq $action } @EFS::CLI::unauthorized_commands ) {
            $self->authrequired(0);
        }

    }

    $self->metadata($metadata);
    $self->operation($operation);
    $self->handle_envvars;

    return 1;

}

##  no critic: LocalizedPunctuationVars
sub handle_envvars {

    my $self = shift;

    my $metadata = $self->metadata;

    my $client_envvars = $metadata->{envvars};

    if ( $client_envvars->{ANSI_COLORS_DISABLED} ) {
        $ENV{ANSI_COLORS_DISABLED} = $client_envvars->{ANSI_COLORS_DISABLED};
    } else {
        delete $ENV{ANSI_COLORS_DISABLED};
    }

    foreach my $var ( keys %{ $client_envvars } ) {
        next unless defined $client_envvars->{$var};
        $ENV{qq{CLIENT_$var}} = $client_envvars->{$var};
    }

    return 1;

}

sub create_logfile {

    my $self  = shift;

    given ( $self->conntype ) {
        when ( q{CLI} ) {
            if ( $self->authrequired ) {
                $self->logfile(EFS::Logfile->new);
            }
        }
        when ( q{RPC} ) {
            if ( $self->metadata->{history}->{parentid} ) {
                $self->logfile(EFS::Logfile->new);
            }
        }
    }

    return 1;

}

sub clone_history {

    my $self = shift;

    return if not $self->isa( q{EFS::Request::RPC} );

    my $metadata = $self->metadata;

    return if not $metadata->{history}->{parentid};

    my $history_rs = EFS::DBI->get_resultset( q{History} );
    my $parent = $history_rs->retrieve_assert( id => $metadata->{history}->{parentid} );

    my $history = $parent->create_child( %{ $metadata->{history}->{values} } );

    $history->set_column( logfile => $self->logfile->path );
    $history->update;

    $self->history_obj($history);

    return 1;

}

sub authenticate_pam_server {

    my $self = shift;

    my $authuser = shift;
    my $authpass = shift;

    my $client = $self->client;

    if ( $authuser eq q{efstestuser} ) {
        if ( $authpass eq q{correct} ) {
            $self->authuser( $authuser );
            return 1;
        } else {
            $client->print( qq{_EFS_INVALID_PASSWORD: $authuser\n} );
            exit 1;
        }
    }

    EFS::Username->new({ name => $authuser });

    if ( not $self->validate_password( $authuser, $authpass ) ) {
        $client->print( qq{_EFS_INVALID_PASSWORD: $authuser\n} );
        exit 1;
    }

    $self->authuser( $authuser );
    return 1;

}

sub validate_password {

    my $self = shift;

    my $authuser = shift;
    my $authpass = shift;

    # XXX: Temporary test environment hack (wpm, 1/25/10)
    if ( $authuser eq q{efsops} and -f q{/etc/efs/test.conf} ) {
        return 1;
    }

    my $ttyname = ttyname( fileno (STDIN) ) || q{/dev/console};

    my $conversation = sub {
        my @results;
        while (@_) {
            my $code    = shift;
            my $message = shift;
            my $answer  = q{};

            if ( $code == PAM_PROMPT_ECHO_ON() ) {
                $answer = $authuser;
            }

            if ( $code == PAM_PROMPT_ECHO_OFF() ) {
                $answer = $authpass;
            }

            push @results, ( PAM_SUCCESS(), $answer );
        }
        push @results, PAM_SUCCESS();
        return @results;
    };

    my $parameter_rs = EFS::DBI->get_resultset(q{Parameter});
    my $service = $parameter_rs->configured( q{pam_service} ) || q{passwd};

    my $pamh = Authen::PAM->new( $service, $authuser, $conversation );

    if ( not ref $pamh ) {
        $self->fatal( qq{Unable to initialize PAM authentication: $pamh\n} );
    }

    $pamh->pam_set_item( PAM_TTY(), $ttyname );

    my $result = $pamh->pam_authenticate;

    if ( $result == PAM_SUCCESS() ) {
        return 1;
    } elsif ( $result == PAM_AUTH_ERR() ) {
        return;
    } else {
        my $error = $pamh->pam_strerror($result);
        $self->fatal( qq{Authentication failure: $error\n} );
    }

    return;

}

sub authenticate_gssapi_server {

    require GSSAPI;
    require GSSAPI::Status;

    my $self = shift;

    my $client = $self->client;

    my $gss_input_token = $client->getline or
        $self->fatal( qq{No GSS input token received from client\n} );

    $gss_input_token = decode_base64( $gss_input_token ) or
        $self->fatal( qq{Unable to decode GSS input token\n} );

    my $context;
    my $gss_output_token;
    my $gss_client_name;
    my $status = GSSAPI::Context::accept(
        $context,
        GSSAPI::GSS_C_NO_CREDENTIAL(),
        $gss_input_token,
        GSSAPI::GSS_C_NO_CHANNEL_BINDINGS(),
        $gss_client_name,
        undef,
        $gss_output_token,
        undef,
        undef,
        undef,
    );
    $status or $self->gss_fatal_server( qq{Unable to authenticate client\n}, $status );

    my $client_name;
    $status = $gss_client_name->display( $client_name );
    $status or $self->gss_fatal_server( qq{Unable to display GSS client name\n}, $status );

    my $client_host = $self->authhost;
    EFS::Syslog->message( q{info}, qq{Authenticated principal $client_name from $client_host} );

    $gss_output_token or $self->fatal( qq{No GSS output token to send back\n} );

    $gss_output_token = encode_base64( $gss_output_token, q{} ) or
        $self->fatal( qq{Unable to encode GSS output token\n} );

    $client->print( qq{$gss_output_token\n} );

    my ($authuser,$realm) = split m{@}msx, $client_name;

    # XXX: How do we validate the realm?

    if ( $authuser =~ m{/}msx ) {
        my ($user,$host) = split m{/}msx, $authuser, 2;
        $self->authuser( $user );
        $self->authhost( $host );
    }
    else {
        $self->authuser( $authuser );
    }

    return 1;

}

sub authenticate_gssapi_client {

    require GSSAPI;
    require GSSAPI::Status;

    my $class      = shift;
    my $connection = shift;
    my $server     = shift;

    my $service_object;
    my $service_name = join q{@}, q{efsd}, $server;
    my $status = GSSAPI::Name->import(
        $service_object,
        $service_name,
        GSSAPI::OID::gss_nt_service_name(),
    );
    $status or $class->gss_fatal_client( qq{Unable to import service name $service_name\n}, $status );

    my $context;
    my $gss_output_unused = q{};
    my $gss_output_token;
    $status = GSSAPI::Context::init(
        $context,
        GSSAPI::GSS_C_NO_CREDENTIAL(),
        $service_object,
        GSSAPI::GSS_C_NO_OID(),
        GSSAPI::GSS_C_MUTUAL_FLAG(),
        0,
        GSSAPI::GSS_C_NO_CHANNEL_BINDINGS(),
        $gss_output_unused,
        undef,
        $gss_output_token,
        undef,
        undef,
    );
    $status or $class->gss_fatal_client( qq{Unable to initialize output context\n}, $status );

    $gss_output_token or $class->gss_fatal_client( qq{No GSS output token to send\n}, $status );

    $gss_output_token = encode_base64( $gss_output_token, q{} ) or
        die qq{Unable to encode GS output token\n};

    $connection->print( qq{$gss_output_token\n} );

    my $gss_input_token = $connection->getline or
        die qq{No GSS inoput token received from server\n};

    $gss_input_token = decode_base64( $gss_input_token ) or
        die qq{Unable to decode GSS input token\n};

    my $gss_input_flags = 0;
    my $gss_input_unused;

    $status = GSSAPI::Context::init(
        $context,
        GSSAPI::GSS_C_NO_CREDENTIAL(),
        $service_object,
        GSSAPI::GSS_C_NO_OID(),
        $gss_input_flags,
        0,
        GSSAPI::GSS_C_NO_CHANNEL_BINDINGS(),
        $gss_input_token,
        undef,
        $gss_input_unused,
        undef,
        undef,
    );
    $status or $class->gss_fatal_client( qq{Unable to initialize input context\n}, $status );

    if ( $ENV{EFS_CLIENT_DEBUG} ) {
        STDOUT->print( qq{[DEBUG] Authenticated via GSSAPI with $service_name\n} );
    }

    return 1;

}

sub gss_fatal_client {
    my $class = shift;
    my ($error,$status) = @_;
    $error .= $class->gss_error($status);
    die $error;
}

sub gss_fatal_server {
    my $self = shift;
    my ($error,$status) = @_;
    $error .= $self->gss_error($status);
    $self->fatal( $error );
}

sub gss_error {

    my $class = shift;
    my $status = shift;
    my $error = q{};

    if ( my @generic = $status->generic_message ) {
        $error .= join(
            qq{\n},
            map { qq{GSS Generic Error: $_} } @generic
        ) . qq{\n};
    }

    if ( my @specific = $status->specific_message ) {
        $error .= join(
            qq{\n},
            map { qq{GSS Specific Error: $_} } @specific
        ) . qq{\n};
    }

    return $error;

}

sub fatal {
    my $self = shift;
    my @errors = @_;
    EFS::Syslog->message( q{err}, @errors );
    $self->client->print( @errors );
    die @errors;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
