package EFS::Util;

use strict;
use warnings;
use English;

use Params::Validate qw( :all );
use POSIX qw( S_ISDIR );
use Regexp::Common;
use File::Basename qw( dirname basename );
use File::Compare;
use File::stat;
use File::Slurp;
use File::Path qw( rmtree );
use Lchown;
use Readonly;
use IO::Dir;

sub GetSubdirs {

    my $class = shift;
    my $dir   = shift;

    my $io = IO::Dir->new( $dir ) or die qq{Unable to opendir $dir: $ERRNO\n};

    my @subdirs = ();

    foreach my $dirent ( $io->read ) {
        next if $dirent ~~ [qw( . .. )];
        next if -l qq{$dir/$dirent};
        next if not -d qq{$dir/$dirent};
        push @subdirs, $dirent;
    }

    return @subdirs;

}

sub GetPathInfo {

    my $class = shift;
    my $path  = shift;

    my $pathinfo = { type => q{unknown} };

    if ( -l $path ) {
        my $readlink = $class->ReadLink( from => $path );
        $pathinfo = {
            type   => q{link},
            target => $readlink,
        };
    }
    elsif ( -d $path ) {
        $pathinfo->{ type } = q{dir};
    }
    elsif ( -f $path ) {
        $pathinfo->{ type } = q{file};
    }

    return $pathinfo;

}

##  no critic: Capitalization  # TODO review
sub MakeDir {

    my $class = shift;

    my (%args) = validate( @_, {
        path_show => { type => SCALAR, optional => 1 },
        path      => { type => SCALAR },
        user      => { type => SCALAR, optional => 1 },
        group     => { type => SCALAR, optional => 1 },
        mode      => { type => SCALAR, default => 02775, optional => 1 },
    });

    my $path      = $args{path};
    my $path_show = $args{path_show} || EFS::DBI->admin_path_show($path);
    my $user      = $args{user};
    my $group     = $args{group};
    my $mode      = $args{mode};

    if ( $EUID == 0 ) {
        $user  ||= q{root};
        $group ||= q{root};
    }
    else {
        if ( $user || $group ) {
            die(
                qq{The user and/or group can only be specified },
                qq{when running as root\n},
            );
        }
    }

    my $changes = 0;

    if ( -l $path ) {
        unlink $path || die qq{Unable to unlink $path: $ERRNO};
        STDOUT->print( qq{Unlinked symlink $path_show\n} );
        $changes++;
    }

    if ( not -d $path ) {
        mkdir( $path, $mode ) || die qq{Unable to mkdir $path: $ERRNO\n};
        STDOUT->print( qq{Created directory $path_show\n} );
        $changes++;
    }

    my $stat = stat( $path )
        or die qq{Unable to stat $path: $ERRNO\n};

    # NOTE: Do NOT lookup root/root via the get*nam() calls, since the
    # root group maps to different numeric IDs on different OSes.

    my $uid = 0;
    my $gid = 0;

    if ( $EUID == 0 ) {

        if ( $user ne q{root} ) {
            $uid = EFS::Username->new({ name => $user })->userid;
        }

        if ( $group ne q{root} ) {
            $gid = EFS::Groupname->new({ name => $group })->groupid;
        }

        if ( $uid != $stat->uid || $gid != $stat->gid ) {
            chown( $uid, $gid, $path ) ||
                die qq{Unable to chown $path: $ERRNO\n};
            STDOUT->print(
                qq{Changed ownership of $path_show to },
                qq{$user/$group\n},
            );
            $changes++;
        }

    }

    $changes += $class->Chmod(
        path => $path,
        stat => $stat,
        mode => $mode,
    );

    return $changes;

}

sub Chmod {

    my $class = shift;

    my (%args) = validate( @_, {
        path => { type => SCALAR },
        stat => { type => OBJECT, optional => 1 },
        mode => { type => SCALAR, default => 02775, optional => 1 },
    });

    my $path = $args{path};
    my $mode = $args{mode};
    my $stat = $args{stat} || stat( $path ) || die qq{Unable to stat $path: $ERRNO\n};

    my $have = sprintf( qq{%lo}, ( $stat->mode & 07777 ) );
    my $want = sprintf( qq{%lo}, $mode );

    if ( $have ne $want ) {
        STDOUT->print( qq{Changing UNIX modes from $have to $want on $path\n} );
        chmod( $mode, $path ) || die qq{Unable to chmod $want $path: $ERRNO\n};
        return 1;
    } else {
        return 0;
    }

}


##  no critic: Capitalization  # TODO review
sub MakeDirHierarchy {

    my $class = shift;

    my (%args) = validate( @_, {
        top      => { type => SCALAR },
        top_show => { type => SCALAR, optional => 1 },
        bottom   => { type => SCALAR },
        user     => { type => SCALAR, default => q{root}, optional => 1 },
        group    => { type => SCALAR, default => q{root}, optional => 1 },
        mode     => { type => SCALAR, default  => 02775, optional => 1 },
    });

    my $changes = 0;

    foreach my $argument (qw( top_show top bottom )) {
        next unless exists $args{$argument};
        $args{$argument} =~ s:/+:/:gx;
        $args{$argument} =~ s:/$::x;
    }

    my $top        = $args{top};
    my $top_show   = $args{top_show} || EFS::DBI->admin_path_show($top);
    my $bottom     = $args{bottom};
    my $user       = $args{user};
    my $group      = $args{group};
    my $mode       = $args{mode};

    if ( $bottom !~ m{^ $top /}msx ) {
        die(
            qq{Invalid directory hierachy specification\n},
            qq{Top ($top) must be a substring of bottom ($bottom)\n},
        );
    }

    my $dir = $bottom;
    my @dirs = ();

    while ( $dir ne $top ) {
        unshift @dirs, $dir;
        $dir = dirname $dir;
    }

    unshift @dirs, $dir;

    foreach my $path ( @dirs ) {

        ( my $path_show = $path ) =~ s{^ $top}{$top_show}msx;

        $changes += $class->MakeDir(
            path_show => $path_show,
            path      => $path,
            user      => $user,
            group     => $group,
            mode      => $mode,
        );

    }

    return $changes;

}

sub RemoveDirHierarchies {

    my $class = shift;

    my (%args) = validate( @_, {
        tops => { type => ARRAYREF },
    });

    foreach my $top ( @{ $args{tops} } ) {
        $class->RemoveDirHierarchy( top => $top );
    }

    return 1;

}

sub RemoveDirHierarchy {

    my $class = shift;

    my (%args) = validate( @_, {
        top      => { type => SCALAR },
        top_show => { type => SCALAR, optional => 1 },
        empty    => { type => SCALAR, optional => 1 },
    });

    my $top      = $args{top};
    my $top_show = $args{top_show} || EFS::DBI->admin_path_show($top);

    if ( not -d $top ) {
        STDOUT->print( qq{No such directory $top_show\n} );
        return 1;
    }

    if ( $args{empty} ) {
        my $io = IO::Dir->new($top) ||
            die qq{Unable to opendir $top: $ERRNO\n};
        if ( grep { not $_ ~~ [qw(. ..)] } $io->read ) {
            die qq{Unable to rmtree $top: not empty\n};
        }
    }


    # Note: F::P::rmtree() doesn't throw proper exceptions for all
    #       errors ... passing in a ref to a scalar (which it will treat
    #       as an array ref), and it will collect all errors there.
    my $error;
    rmtree( $top, { error => \$error } );
    if ( @$error ) {
        my $message = q{};
        for my $diag ( @$error ) {
            my ($file, $err) = each %$diag;
            $message .= ( $file ? $file : q{general error} )
                . qq{: $error\n};
        }
        die qq{Unable to rmtree $top: $message\n};
    }

    STDOUT->print( qq{Removed directory $top_show\n} );
    return 1;

}

sub ReadLink {

    my $class = shift;

    my (%args) = validate( @_, {
        from => { type => SCALAR },
    });

    my $from = $args{from};

    my $readlink = readlink $from ||
        die qq{Unable to readlink $from: $ERRNO\n};

    return $readlink;

}

##  no critic: Capitalization  # TODO review
sub RemoveLink {

    my $class = shift;

    my (%args) = validate( @_, {
        from      => { type => SCALAR },
        from_show => { type => SCALAR, optional => 1 },
    });

    my $from      = $args{from};
    my $from_show = $args{from_show} || EFS::DBI->admin_path_show($from);

    if ( -l $from ) {
        unlink $from || die qq{Unable to remove symlink $from: $ERRNO\n};
        STDOUT->print( qq{Removed symlink $from_show\n} );
    }
    else {
        STDOUT->print( qq{No such symlink $from_show\n} );
    }

    return 1;

}

sub RemoveFile {

    my $class = shift;

    my (%args) = validate( @_, {
        file => { type => SCALAR },
    });

    my $file      = $args{file};

    if ( -f $file ) {
        unlink $file || die qq{Unable to remove file $file: $ERRNO\n};
    }

    return 1;

}

##  no critic: Capitalization  # TODO review
sub MakeLink {

    my $class = shift;

    my (%args) = validate( @_, {
        from      => { type => SCALAR },
        to        => { type => SCALAR },
        query     => { type => BOOLEAN, optional => 1 },
        from_show => { type => SCALAR, optional => 1 },
    });

    foreach my $argument (qw( from_show from to )) {
        next unless exists $args{$argument};
        $args{$argument} =~ s:/+:/:gx;
        $args{$argument} =~ s:/$::x;
    }

    my $from      = $args{from};
    my $to        = $args{to};
    my $from_show = $args{from_show} || EFS::DBI->admin_path_show($from);

    if ( $from eq $to ) {
        die qq{Invalid from and to specified: they match\n};
    }

    if ( -l $from ) {

        my $readlink = readlink( $from ) ||
            die qq{Unable to readlink $from: $ERRNO\n};

        if ( $readlink eq $to ) {
            return 0;
        }

        my $templink = $from . qq{-$PID-} . time;

        symlink( $to, $templink ) ||
            die qq{Unable to create symlink $templink: $ERRNO\n};

        rename( $templink, $from ) ||
            die qq{Unable to rename $templink -> $from: $ERRNO\n};

        STDOUT->print( qq{Changed symlink $from_show from $readlink to $to\n} );

        return 1;

    }
    elsif ( -f $from || -d $from ) {
        die(
            qq{Unable to create or update symlink $from\n},
            q{It already exists as a }
                . ( -d $from ? q{directory} : q{file} ) . qq{\n}
        );
    }
    else {

        symlink( $to, $from ) ||
            die qq{Unable to create symlink $from: $ERRNO\n};

        STDOUT->print( qq{Created symlink $from_show -> $to\n} );

        return 1;

    }

}

sub RenameDir {

    my $class = shift;

    my (%args) = validate( @_, {
        from  => { type     => SCALAR },
        to    => { type     => SCALAR },
        quiet => { type     => SCALAR,
                   optional => 1 },
    });

    my $from      = $args{from};
    my $from_show = EFS::DBI->admin_path_show($from);
    my $to        = $args{to};
    my $to_show   = EFS::DBI->admin_path_show($to);

    if ( -l $to ) {
        $class->RemoveLink( from => $to );
    }
    elsif ( -d $to ) {
        die qq{Unable to rename $from to $to: target is a directory\n};
    }

    rename( $from, $to ) ||
        die qq{Unable to rename $from to $to: $ERRNO\n};

    STDOUT->print( qq{Renamed $from_show to $to_show\n} );

    return 1;

}

##  no critic: Capitalization  # TODO review
sub WaitDir {

    my $class = shift;

    my (%args) = validate( @_, {
        directory => { type => SCALAR },
        retries   => { type => SCALAR, default => 60, optional => 1 },
        sleep     => { type => SCALAR, default => 5, optional => 1 },
    });

    foreach my $argument ( qw( sleep retries ) ) {
        if ( not ( $RE{num}{int}->matches( $args{$argument} ) &&
                       $args{$argument} > 0 ) ) {
            die(
                qq{Invalid $argument value '$args{$argument}'. },
                qq{Must be a positive integer\n}
            );
        }
    }

    my $directory = $args{directory};
    my $sleep     = $args{sleep};
    my $retries   = $args{retries};

    my $retry_count = $retries;

    my $parentdir = dirname $directory;

    while ( $retry_count ) {

        if ( $parentdir ne q{/} ) {

            # NFS cache workaround.  Stat the parent dir to avoid
            # stale cache entries.
            $class->WaitDir(
                directory => $parentdir,
                retries   => $retries,
                sleep     => $sleep,
            );
        }

        if ( IO::Dir->new($directory) ) {
            return 1;
        }

        sleep $sleep;
        $retry_count--;

    }

    my $retries_used = $retries - $retry_count;

    if ( not -d $directory ) {
        die qq{Directory $directory never appeared\n};
    }

    return 1;

}

# XXX: We need to add support for a "mode" argument, similar to MakeDir

##  no critic: Capitalization  # TODO review
sub Chown {

    ##  no critic (MagicNumbers)

    my $class = shift;

    my (%args) = validate( @_, {
        path          => { type => SCALAR },
        user          => { type => SCALAR },
        group         => { type => SCALAR },
        lockgroup     => { type => SCALAR },
        uid           => { type => SCALAR, optional => 1 },
        gid           => { type => SCALAR, optional => 1 },
        lockgid       => { type => SCALAR, optional => 1 },
        lockmode      => { type => SCALAR, default => q{none}, optional => 1 },
        verbose       => { type => BOOLEAN, optional => 1 },
        noworldaccess => { type => SCALAR | OBJECT | UNDEF, optional => 1 },
    });

    my $path          = $args{path};
    my $user          = $args{user};
    my $group         = $args{group};
    my $lockgroup     = $args{lockgroup};
    my $lockmode      = $args{lockmode};
    my $verbose       = $args{verbose};
    my $noworldaccess = $args{noworldaccess};

    my $path_show = EFS::DBI->admin_path_show($path);

    my $uid     = 0;
    my $gid     = 0;
    my $lockgid = 0;
    my $lockuid = 0;

    if ( exists $args{uid} ) {
        $uid = $args{uid};
    }
    else {
        if ( $user ne q{root} ) {
            $uid = EFS::Username->new({ name => $user })->userid;
        }
    }

    if ( exists $args{gid} ) {
        $gid = $args{gid};
    }
    else {
        if ( $group ne q{root} ) {
            $gid = EFS::Groupname->new({ name => $group })->groupid;
        }
    }

    if ( exists $args{lockgid} ) {
        $lockgid = $args{lockgid};
    }
    else {
        if ( $lockgroup ne q{root} ) {
            $lockgid = EFS::Groupname->new({ name => $lockgroup })->groupid;
        }
    }

    my $stat;
    my $path_is_symlink = 0;

    if ( -l $path ) {
        $stat = lstat( $path ) or
            die(qq{Unable to lstat $path: $ERRNO\n});
        $path_is_symlink = 1;
    }
    else {
        $stat = stat( $path ) or
            die qq{Unable to stat $path: $ERRNO\n};
    }

    my $changed = 0;

    if ( $uid != $stat->uid || $gid != $stat->gid ) {

        if (   $stat->uid == $lockuid
            && $stat->gid == $lockgid
            && $lockmode ne q{unlock} ) {

            if ( $verbose ) {
                STDOUT->print( qq{Skipping locked down path $path_show\n} );
            }

        }
        else {

            if ( $path_is_symlink ) {
                lchown( $uid, $gid, $path ) ||
                    die qq{Unable to lchown $path: $ERRNO\n};
            }
            else {
                chown( $uid, $gid, $path ) ||
                    die qq{Unable to chown $path: $ERRNO\n};
            }
            if ( $verbose ) {
                STDOUT->print(
                    qq{Changed ownership of $path_show to $user/$group\n}
                );
            }
            $changed = 1;

        }

    }

    if ( $path_is_symlink ) {
        return $changed;
    }

    if ( $lockmode eq q{unlock} ) {

        if ( S_ISDIR( $stat->mode ) ) {

            # Case: directory unlock

            if ( not ( $stat->mode & 02000 ) ) {
                chmod( 07777 & ( $stat->mode | 02000 ), $path ) ||
                    die qq{Unable to chmod g+s $path: $ERRNO\n};
                if ( $verbose ) {
                    STDOUT->print( qq{Set group sticky bit on $path_show\n} );
                }
                $changed = 1;
            }

            if ( not ( $stat->mode & 020 ) ) {
                chmod( 07777 & ( $stat->mode | 020 ), $path ) ||
                    die qq{Unable to chmod g+w $path: $ERRNO\n};
                if ( $verbose ) {
                    STDOUT->print( qq{Set group write bit on $path_show\n} );
                }
                $changed = 1;
            }

        }
        else {

            # Case: file unlock

            if ( ( $stat->mode & 0200 ) && ( not ( $stat->mode & 020 ) ) ) {
                chmod( 07777 & ( $stat->mode | 020 ), $path ) ||
                    die qq{Unable to chmod g+w $path: $ERRNO\n};
                if ( $verbose ) {
                    STDOUT->print( qq{Set group write bit on $path_show\n} );
                }
                $changed = 1;
            }

        }

    }
    else {

        # Case: lock

        # If the group is something OTHER than root, chmod g-w
        if ( $lockgid != 0 && $stat->mode & 020 ) {
            chmod( 07777 & ( $stat->mode & ( ~020 ) ), $path ) ||
                die qq{Unable to chmod g-w $path: $ERRNO\n};
            if ( $verbose ) {
                STDOUT->print( qq{Removed group write bit on $path_show\n} );
            }
            $changed = 1;
        }

    }

    if ( $noworldaccess ) {

        if ( $stat->mode & 07 ) {
            chmod( 07777 & ( $stat->mode & ( ~07 ) ), $path ) ||
                die qq{Unable to chmod o-rwx $path: $ERRNO\n};
            if ( $verbose ) {
                STDOUT->print( qq{Removed world access bits on $path_show\n} );
            }
            $changed = 1;
        }

    }
    else {

        if ( S_ISDIR( $stat->mode ) ) {
            if ( not ( $stat->mode & 05 ) ) {
                chmod( 07777 & ( $stat->mode | 05 ), $path ) ||
                    die qq{Unable to chmod o+rx $path: $ERRNO\n};
                if ( $verbose ) {
                    STDOUT->print( qq{Set world access bits on $path_show\n} );
                }
                $changed = 1;
            }
        }

        if ( $stat->mode & 02 ) {
            chmod( 07777 & ( $stat->mode & ( ~02 ) ), $path ) ||
                die qq{Unable to chmod o-w $path: $ERRNO\n};
            if ( $verbose ) {
                STDOUT->print( qq{Removed world write bit on $path_show\n} );
            }
            $changed = 1;
        }

    }

    if ( -l $path ) {
        return $changed;
    }

    if ( not -d $path ) {
        return $changed;
    }

    my $io = IO::Dir->new( $path ) ||
        die qq{Unable to opendir $path: $ERRNO\n};

    foreach my $dirent ( $io->read ) {

        next if $dirent ~~[qw(. ..)];

        $changed += EFS::Util->Chown(
            path          => qq{$path/$dirent},
            user          => $user,
            group         => $group,
            lockgroup     => $lockgroup,
            uid           => $uid,
            gid           => $gid,
            lockgid       => $lockgid,
            verbose       => $verbose,
            lockmode      => $lockmode,
            noworldaccess => $noworldaccess,
        );

    }

    return $changed;

}

sub CreateLinkedDirectory {

    my $class = shift;

    my (%args) = validate( @_, {
        from      => { type => SCALAR },
        from_show => { type => SCALAR, optional => 1 },
        to        => { type => SCALAR },
        to_show   => { type => SCALAR, optional => 1 },
    });

    my $from      = $args{from};
    my $from_show = $args{from_show} || EFS::DBI->admin_path_show($from);
    my $to        = $args{to};
    my $to_show   = $args{to_show} || EFS::DBI->admin_path_show($to);

    my $to_io = IO::Dir->new( $to ) ||
        die qq{Unable to opendir $to: $ERRNO\n};

    my %want = ();

    foreach my $name ( $to_io->read ) {
        next if $name ~~ [qw(. ..)];
        $want{ $name } = qq{$to/$name};
    }

    my $from_io = IO::Dir->new( $from ) ||
        die qq{Unable to opendir $from: $ERRNO\n};

    my %have = ();

    foreach my $name ( $from_io->read ) {

        next if $name ~~ [qw(. ..)];

        if ( -l qq{$from/$name} ) {
            $have{ $name } =
                EFS::Util->ReadLink( from => qq{$from/$name} );
        }
        else {
            die qq{Not a symlink: $from/$name\n};
        }

    }

    foreach my $name ( sort keys %want ) {

        if ( $have{$name} and $have{$name} eq $want{$name} ) {
            next;
        }

        EFS::Util->MakeLink(
            from => qq{$from/$name},
            to   => $want{$name},
        );

    }

    foreach my $name ( sort keys %have ) {
        if ( not $want{$name} ) {
            EFS::Util->RemoveLink( from => qq{$from/$name} );
        }
    }

    return 1;

}

sub CloneDirectory {

    my $class = shift;

    my (%args) = validate( @_, {
        from      => { type     => SCALAR },
        to        => { type     => SCALAR },
        replace   => { type     => HASHREF,
                       optional => 1 },
        from_show => { type     => SCALAR,
                       optional => 1 },
        to_show   => { type     => SCALAR,
                       optional => 1 },
    });

    my $from      = $args{from};
    my $from_show = $args{from_show} || EFS::DBI->admin_path_show($from);
    my $to        = $args{to};
    my $to_show   = $args{to_show}   || EFS::DBI->admin_path_show($to);

    if ( not -d $from ) {
        die qq{No such directory: $from\n};
    }

    EFS::Rsync->new({
        src_path  => $from,
        dest_path => $to,
    });

    if ( $args{replace} ) {
        my $filetree = EFS::FileTree->new({ pathname => $to });
        $filetree->replace_paths( %{ $args{replace} } );
    }

    STDOUT->print( qq{Cloned directory $from_show to $to_show\n} );

    return 1;

}

#
# This is one of those hacks that SHOULD be unnecessary, but since we're
# suffering with the asynchronicity of the NFS client on some flaky platforms
# (eg. AIX), this is proving an essential workaround.
#
sub WriteFileRetry {

    my $class  = shift;
    my (%args) = @_; 

    my $pathname = $args{pathname};
    my $contents = $args{contents};

    my $retries     = $args{retries}     || 6;
    my $retry_sleep = $args{retry_sleep} || 10;

#
# NOTE: This code does NOT WORK!!!  Inside the catch block, $_ is
# uninitialized EVERY time we hit it.  I can not reproduce this in a
# simple test case.  Something obscure is breaking this.
#    
#  WRITEFILE: {
#         try {
#             write_file( $pathname, $contents );
#         } catch {
#             die $_ if not m{Text busy}ms;
#             die $_ if not $retries--;
#             warn $_;
#             sleep $retry_sleep;
#             redo WRITEFILE;
#         };
#    }

  WRITEFILE: {
        eval { write_file( $pathname, $contents ); };
        if ( $EVAL_ERROR ) {
            die $EVAL_ERROR if not $EVAL_ERROR =~ m{Text busy}ms;
            die $EVAL_ERROR if not $retries--;
            warn $EVAL_ERROR;
            sleep $retry_sleep;
            redo WRITEFILE;
        }
    }

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
