#!/usr/bin/env perl
use strict;
use warnings;
use English;

use IO::File;

use lib q{test/lib};
use EFS::Test::Config;

use Test::More;

use POSIX qw( setuid setgid );
use Text::Password::Pronounceable;
use IO::File;
use DBI;

my $config = EFS::Test::Config->new;

my $ldc_efsserver = $config->get_value( q{LocalDevCell},  q{efsserver} );
my $rdc_efsserver = $config->get_value( q{RemoteDevCell}, q{efsserver} );
my $sshcmd        = $config->get_value( q{Global},        q{sshcmd} );

my $port = $config->get_value( q{Database}, q{port} );
my $user = $config->get_value( q{Database}, q{user} );

if ( $config->get_value( q{Database}, q{manual} ) ) {
    plan skip_all => q{Automated database setup disabled};
}

my $connect = $config->get_value( q{Database}, q{connect} );

if ( defined $connect && $connect =~ m{oracle}imsx ) {
    plan skip_all => q{Automated database setup only works for MySQL};
}

# Create /var/efs/db directory structure first

if ( -d q{/var/efs/db} ) {
    die(
        qq{Database directory (/var/efs/db) exists.\n},
        qq{Remove this first (test/99_shutdown/95_dbserver.t)\n}
    );
}

my @getpwnam = getpwnam($user) or
    die qq{Unable to query $user: $ERRNO\n};

my $uid = $getpwnam[2];
my $gid = $getpwnam[3];

my @localdirs = (
    [ q{/var/efs/db}, 0755 ],
    [ q{/var/efs/db/data}, 0755 ],
    [ q{/var/efs/db/logs}, 0755 ],
    [ q{/var/efs/db/passwd}, 0700 ],
);

foreach my $ref (@localdirs) {

    my ($dir,$mode) = @{ $ref };

    mkdir( $dir, $mode ) || die qq{Unable to mkdir $dir: $ERRNO\n};
    ok( 1, qq{mkdir $dir} );

    chown( $uid, $gid, $dir ) || die qq{Unable to chown $dir: $ERRNO\n};
    ok( 1, qq{chown $dir} );

}

my %passwd_of = (
    root      => generate_password( q{/var/efs/db/passwd/root}, $uid, $gid ),
    efs_dbo   => generate_password( q{/var/efs/db/passwd/efs_dbo}, $uid, $gid ),
    efs_write => generate_password( q{/var/efs/db/passwd/efs_write}, $uid, $gid ),
);

$connect = qq{DBI:mysql:database=efs;host=$ldc_efsserver;port=$port};

# Update test.conf

my $save_config = EFS::Test::Config->new( { noexpand => 1 } );

$save_config->set_value( q{Database}, q{connect}, $connect );
$save_config->set_value( q{Database}, q{dbouser}, q{efs_dbo} );
$save_config->set_value( q{Database}, q{dbopasswd}, $passwd_of{efs_dbo} );
$save_config->set_value( q{Database}, q{rwuser}, q{efs_write} );
$save_config->set_value( q{Database}, q{rwpasswd}, $passwd_of{efs_write} );
$save_config->set_value( q{Database}, q{rouser}, q{efs_read} );
$save_config->set_value( q{Database}, q{ropasswd}, q{efs_read} );
$save_config->delete_key( q{Database}, q{prefix} );

$save_config->save_file ||
    die qq{Unable to save test config file: $ERRNO\n};

ok( 1, q{Saved test config file} );

my $mycnf = q{};

foreach my $file ( qw( /etc/my.cnf /etc/mysql/my.cnf ) ) {
    next if not -f $file;
    $mycnf = $file;
    last;
}

if ( not $mycnf ) {
    $mycnf = -d q{/etc/mysql} ? q{/etc/mysql/my.cnf} : q{/etc/my.cnf};
}

if ( -l $mycnf or -f $mycnf ) {

    my $mycnforig = qq{$mycnf.orig};

    if ( not -l $mycnf ) {
        rename( $mycnf, $mycnforig ) ||
            die qq{Unable to rename $mycnf -> $mycnforig: $ERRNO\n};
    } else {
        unlink $mycnf ||
            die qq{Unable to unlink $mycnf: $ERRNO\n};
    }

}

my $conffile = q{/var/efs/db/my.cnf};

symlink( $conffile, $mycnf ) ||
    die qq{Unable to symlink $mycnf -> $conffile: $ERRNO\n};

ok( 1, qq{Made $mycnf a symlink} );

# Become the database user

setgid( $gid ) or die qq{Unable to setgid: $ERRNO\n};
setuid( $uid ) or die qq{Unable to setuid: $ERRNO\n};

ok( 1, q{Changed uid/gid to efsops} );

# Do everything else as the user

# Create mysql config files

my $confio = IO::File->new(qq{>$conffile}) ||
    die qq{Unable to write to $conffile: $ERRNO\n};

$confio->print(<<"END_OF_CONFFILE");
#
# EFS MySQL Config
#
[client]
port                    = $port
socket                  = /var/efs/db/mysql.sock

[mysqld]
user                    = efsops
port                    = $port
socket                  = /var/efs/db/mysql.sock
datadir                 = /var/efs/db/data
# Only works with 5.1.12 or later
# CentOS 5.4 still has 5.0
#general_log             = /var/efs/db/logs/mysql.log
log_error               = /var/efs/db/logs/error.log
pid_file                = /var/run/efsd-mysql.pid
innodb                  = FORCE
character_set_server    = latin1
collation_server        = latin1_general_cs
END_OF_CONFFILE

$confio->close ||
    die qq{Unable to close $conffile: $ERRNO\n};

# Install the db and start mysqld

my $output = qx{mysql_install_db 2>&1};
$CHILD_ERROR and die qq{Unable to install db\n$output\n};

ok( 1, q{mysql_install_db} );

system(qq{mysqld_safe < /dev/null >/var/tmp/mysqld_safe.out.$PID 2>&1 &});
if ( $CHILD_ERROR ) {
    my $output = qx{cat /var/tmp/mysqld_safe.out.$PID};
    die qq{Unable to start mysqld\n$output\n};
}

# This ugliness is what /etc/init.d/mysqld does.  Blegh...

my $retries  = 30;
my $alive    = 0;
my $response = q{};

while ( $retries-- ) {
    $response = qx{mysqladmin --user=root ping 2>&1};
    if ( $response =~ m{mysqld \s+ is \s+ alive}msx ) {
        $alive = 1;
        last;
    }
    sleep 2;
}

if ( not $alive ) {
    die qq{Unable to start database.\n$response\n};
}

ok( 1, q{mysqld_safe} );

$output =
    qx{echo 'create database efs;' | mysql -u root};
$CHILD_ERROR and die qq{Unable to create efs database\n$output\n};

ok( 1, q{Created efs database} );

# Create users

$output =
    qx{mysqladmin -u root password $passwd_of{root} 2>&1};
$CHILD_ERROR and die qq{Unable to set root password:\n$output\n};

ok( 1, q{Created new passwords} );

my $sqlfile = qq{/tmp/efs_sql.$PID};

my $tmp = IO::File->new(qq{>$sqlfile}) ||
    die qq{Unable to open $sqlfile: $ERRNO\n};

$tmp->print(<<"END_OF_SQLFILE");
USE 'mysql';
CREATE USER 'efs_dbo'\@'%' IDENTIFIED BY '$passwd_of{efs_dbo}';
GRANT ALL ON efs.* TO 'efs_dbo'\@'%' WITH GRANT OPTION;
CREATE USER 'efs_write'\@'%' IDENTIFIED BY '$passwd_of{efs_write}';
GRANT SELECT, UPDATE, INSERT, DELETE ON efs.* TO 'efs_write'\@'%';
CREATE USER 'efs_read'\@'%' IDENTIFIED BY 'efs_read';
GRANT SELECT ON efs.* TO 'efs_read'\@'%';
FLUSH PRIVILEGES;
SELECT User,Host from mysql.user;
SHOW GRANTS for 'efs_dbo'\@'%';
SHOW GRANTS for 'efs_write'\@'%';
SHOW GRANTS for 'efs_read'\@'%';
END_OF_SQLFILE

$output =
    qx{mysql --user=root --password=$passwd_of{root} < $sqlfile 2>&1};
$CHILD_ERROR and die qq{Unable to create efs_write user\n$output\n};

ok( 1, q{Updated user table} );

my $dbh = DBI->connect( $connect, q{efs_write}, $passwd_of{efs_write} );
ok( $dbh, q{Connected to database via DBI} );

done_testing();

exit 0;

sub generate_password {

    my ($file,$uid,$gid) = @_;

    my $io = IO::File->new(qq{>$file}) ||
        die qq{Unable to open $file: $ERRNO\n};

    my $password = Text::Password::Pronounceable->generate( 20, 30 );

    $io->print(qq{$password\n}) ||
        die qq{Unable to write to $file: $ERRNO\n};

    $io->close ||
        die qq{Unable to close $file: $ERRNO\n};

    chown( $uid, $gid, $file ) ||
        die qq{Unable to chown $file: $ERRNO\n};

    return $password;

}

1;

