#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 38;

use blib;
use EFS;

use lib q{test/lib};
use EFS::Test;

## no critic: RequireExtendedFormatting

my $test   = EFS::Test->new;
my $config = $test->get_config;

my $username  = $test->get_primary_username;
my $groupname = $test->get_primary_groupname;

my $ldc_name  = $config->get_value( q{LocalDevCell}, q{name} );

my @setup_one = (
    [ 0, q{efs display metaproj testmeta} ],
    [ 1, q{efs display project testmeta testauto} ],
    [ 1, qq{efs display userrights testmeta auto $username} ],
    [ 0, qq{efs create project testmeta testauto $ldc_name} ],
    [ 0, q{efs create release testmeta testauto bogusrelease} ],
    [ 0, q{efs create releasealias testmeta testauto bogusalias bogusrelease} ],
);

foreach my $setup ( @setup_one ) {
    my ($not,$command) = @{ $setup };
    $test->efs_test_command(
        name    => qq{(setup, phase one) $command},
        not     => $not,
        command => $command,
        bail    => 1,
    );
}

$test->efs_test_command(
    command => q{efs create autorelease testmeta testauto},
    not     => 1,
    expect  => [
        q{Access denied.},
        q{The 'auto' role is required to create autorelease},
        q{Exit status from efs server not received.  Assuming failure...},
    ],
);

$test->efs_test_command(
    command => qq{efs define userrights testmeta auto $username},
);

$test->efs_test_command(
    command => q{efs create autorelease testmeta testauto},
    not     => 1,
    expect  => [
        q{The following releases are not autorelease compliant:},
        q{testmeta/testauto/bogusrelease},
        q{The following release aliases are not autorelease compliant:},
        q{testmeta/testauto/bogusalias},
    ],
);

my @setup_two = (
    q{efs destroy releasealias testmeta testauto bogusalias},
    q{efs setstage release testmeta testauto bogusrelease rejected},
    q{efs destroy release testmeta testauto bogusrelease},
    q{efs purge release testmeta testauto bogusrelease --now},
);

foreach my $setup ( @setup_two ) {
    my $run_as = $setup =~ m{--now}ms ? q{efsops} : q{primary_user};
    $test->efs_test_command(
        run_as  => $run_as,
        name    => qq{(setup, phase two) $setup},
        command => $setup,
    );
}

$test->efs_test_command(
    command => q{efs dist autorelease testmeta testauto},
    not     => 1,
    expect  => q{No such releasealias 'testmeta/testauto/next'},
);

my $project_dir = q{/efs/dev/testmeta/testauto};

$test->efs_test_command(
    command => q{efs create autorelease testmeta testauto},
    expect  => [
        qr{Created directory $project_dir/\d{8}-\d{6}}ms,
        qr{Created directory $project_dir/\d{8}-\d{6}/install}ms,
        qr{Created directory $project_dir/\d{8}-\d{6}/build}ms,
        qr{Changed ownership of $project_dir/\d{8}-\d{6}/build to $username/$groupname}ms,
        qr{Created directory $project_dir/\d{8}-\d{6}/src}ms,
        qr{Changed ownership of $project_dir/\d{8}-\d{6}/src to $username/$groupname}ms,
        qr{Creating dist2dev symlink for testmeta/testauto/\d{8}-\d{6}}ms,
        qr{Created symlink $ldc_name:/efs/dist/testmeta/testauto/\d{8}-\d{6} -> $project_dir/\d{8}-\d{6}/install}ms,
        qr{Created release testmeta/testauto/\d{8}-\d{6}}ms,
        qr{Created directory $ldc_name:$project_dir/\d{8}-\d{6}/install/common}ms,
        qr{Changed ownership of $ldc_name:$project_dir/\d{8}-\d{6}/install/common to $username/$groupname}ms,
        qr{Created install $project_dir/\d{8}-\d{6}/install/common}ms,
        qr{Created symlink $project_dir/next -> \d{8}-\d{6}}ms,
        qr{Created releasealias testmeta/testauto/next -> \d{8}-\d{6}}ms,
        qr{Created autorelease testmeta/testauto/\d{8}-\d{6}}ms,
    ],
);

my $project_rs = EFS::DBI->get_resultset( q{Project} );
my $project = $project_rs->retrieve_bynames_assert(
    metaproj => q{testmeta},
    name     => q{testauto},
);

my ($autorelease) = $project->releases;

$test->efs_test_command(
    command => q{efs create autorelease testmeta testauto},
    filter  => qr{Created autorelease}ms,
    expect  => qr{Created autorelease $autorelease}ms,
);

$test->efs_test_command(
    command => q{efs revert autorelease testmeta testauto},
    not     => 1,
    expect  => q{Unable to determine previous release to revert},
);

$test->efs_test_command(
    command => q{efs display project testmeta testauto},
    expect  => [
        qr{Metaproj Project Cell}ms,
        qr{=+}ms,
        qr{testmeta testauto $ldc_name}ms,
        qr{Project Releasealias Target}ms,
        qr{=+}ms,
        qr{testauto next \d{8}-\d{6}}ms,
        qr{Active Releases}ms,
        qr{Metaproj Project Release}ms,
        qr{=+}ms,
        qr{testmeta testauto \d{8}-\d{6} dev}ms,
    ],
);

$test->efs_test_command(
    command => q{touch /efs/dev/testmeta/testauto/next/install/common/foo},
);

$test->efs_test_command(
    command => q{efs dist autorelease testmeta testauto},
    filter  => qr{Successfully distributed release}ms,
    expect  => [
        qr{Successfully distributed release}ms,
        qr{Successfully distributed release}ms,
        qr{Successfully distributed releasealias testmeta/testauto/current}ms,
    ],
);

$test->efs_test_command(
    command => q{efs display project testmeta testauto},
    expect  => [
        qr{Metaproj Project Cell}ms,
        qr{=+}ms,
        qr{testmeta testauto $ldc_name}ms,
        qr{Project Releasealias Target}ms,
        qr{=+}ms,
        qr{testauto current \d{8}-\d{6}}ms,
        qr{Active Releases}ms,
        qr{Metaproj Project Release}ms,
        qr{=+}ms,
        qr{testmeta testauto \d{8}-\d{6} prod}ms,
    ],
);

$test->efs_test_command(
    command => q{efs create autorelease testmeta testauto},
);

$test->efs_test_command(
    command => q{touch /efs/dev/testmeta/testauto/next/install/common/foo},
);

$test->efs_test_command(
    command => q{efs dist autorelease testmeta testauto},
);

$test->efs_test_command(
    command => q{efs display project testmeta testauto},
    expect  => [
        qr{Metaproj Project Cell}ms,
        qr{=+}ms,
        qr{testmeta testauto $ldc_name}ms,
        qr{Project Releasealias Target}ms,
        qr{=+}ms,
        qr{testauto current \d{8}-\d{6}}ms,
        qr{Active Releases}ms,
        qr{Metaproj Project Release}ms,
        qr{=+}ms,
        qr{testmeta testauto \d{8}-\d{6} prod}ms,
        qr{Destroyed Releases}ms,
        qr{Metaproj Project Release Stage}ms,
        qr{=+}ms,
        qr{testmeta testauto \d{8}-\d{6} deprecated yes}ms,
    ],
);

$test->efs_test_command(
    command => q{efs revert autorelease testmeta testauto},
    filter  => qr{Reverted}ms,
    expect  => qr{Reverted to autorelease testmeta/testauto/\d{8}-\d{6}}ms,
);

$test->efs_test_command(
    command => q{efs revert autorelease testmeta testauto},
    not     => 1,
    expect  => q{Unable to determine previous release to revert},
);

my @destroyed = $project->destroyed_releases;

$test->efs_test_command(
    name    => q{Only one destroyed release},
    command => $#destroyed == 0 ? q{true} : q{false},
);

my $revert_name = $destroyed[0]->name;

$test->efs_test_command(
    command => qq{efs revert autorelease testmeta testauto $revert_name},
    not     => 1,
    expect  => [
        q{Unable to revert to a previously reverted autorelease},
        qq{To revert to testmeta/testauto/$revert_name, use --force},
    ],
);

$test->efs_test_command(
    command => qq{efs revert autorelease testmeta testauto $revert_name --force},
    filter  => qr{Reverted}ms,
    expect  => qr{Reverted to autorelease testmeta/testauto/\d{8}-\d{6}}ms,
);

$test->efs_test_command(
    command => q{efs history project testmeta testauto --count 19},
    expect  => [
        qr{Status Authuser Request}ms,
        qr{-+}ms,
        qr{1 $username create project testmeta testauto}ms,
        qr{1 $username create release testmeta testauto bogusrelease}ms,
        qr{1 $username create releasealias testmeta testauto bogusalias bogusrelease}ms,
        qr{0 $username create autorelease testmeta testauto}ms,
        qr{1 $username destroy releasealias testmeta testauto bogusalias}ms,
        qr{1 $username setstage release testmeta testauto bogusrelease rejected}ms,
        qr{1 $username destroy release testmeta testauto bogusrelease}ms,
        qr{1 efsops purge release testmeta testauto bogusrelease --now}ms,
        qr{0 $username dist autorelease testmeta testauto}ms,
        qr{1 $username create autorelease testmeta testauto}ms,
        qr{1 $username create autorelease testmeta testauto}ms,
        qr{0 $username revert autorelease testmeta testauto}ms,
        qr{1 $username dist autorelease testmeta testauto}ms,
        qr{1 $username create autorelease testmeta testauto}ms,
        qr{1 $username dist autorelease testmeta testauto}ms,
        qr{1 $username revert autorelease testmeta testauto}ms,
        qr{0 $username revert autorelease testmeta testauto}ms,
        qr{0 $username revert autorelease testmeta testauto}ms,
        qr{1 $username revert autorelease testmeta testauto}ms,
    ],
);

my @teardown = (
    qq{efs undef userrights testmeta auto $username},
    q{efs destroy releasealias testmeta testauto current},
);

foreach my $destroyed ( $project->destroyed_releases ) {
    my $name = $destroyed->name;
    push @teardown, qq{efs purge release testmeta testauto $name --now};
}

foreach my $active ( $project->active_releases ) {
    my $name = $active->name;
    push @teardown, (
        qq{efs setstage release testmeta testauto $name deprecated},
        qq{efs destroy release testmeta testauto $name},
        qq{efs purge release testmeta testauto $name --now},
    );
}

push @teardown, q{efs destroy project testmeta testauto};

foreach my $teardown ( @teardown ) {
    my $run_as = $teardown =~ m{--now}msx ? q{efsops} : q{primary_user};
    $test->efs_test_command(
        run_as  => $run_as,
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;
