#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 32;

use lib q{test/lib};
use EFS::Test;

my $test     = EFS::Test->new;
my $config   = $test->get_config;

my $username  = $test->get_primary_username;
my $groupname = $test->get_primary_groupname;

my $root_domain = $config->get_value( q{Global}, q{root_domain} );
my $ownerlist = join q{@}, $username, $root_domain;

my $release = q{testmeta/testproj/testrel};

my @setup = (
    q{efs display metaproj testmeta},
    q{efs display project testmeta testproj},
    q{efs create release testmeta testproj testrel},
    q{efs create install testmeta testproj testrel common},
    qq{touch /efs/dev/$release/install/common/foo},
    qq{efs set_notify_list metaproj testmeta $ownerlist},
    q{efs set_noworldaccess metaproj testmeta 1},
    qq{efs set_distgroup metaproj testmeta $groupname},
    qq{efs set_notify_list project testmeta testproj $ownerlist},
    q{efs checkpoint release testmeta testproj testrel},
);

foreach my $setup (@setup) {
    $test->efs_test_command(
        name    => qq{(setup) $setup},
        command => $setup,
        bail    => 1,
    );
}

$test->efs_test_file(
    name  => q{Verify correct ownership/mode when using restricted mode},
    file  => qq{/efs/dev/$release/install/common/foo},
    user  => q{root},
    group => $groupname,
    mode  => 0640,
    bail  => 1,
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => q{efs display feature disable_restricted_access},
    expect  => q{No such feature 'disable_restricted_access'},
    not     => 1,
    bail    => 1,
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => q{efs define feature disable_restricted_access 1},
    expect  => q{Defined feature 'disable_restricted_access'},
);

# Verify that set_noworldaccess, set_distgroup and clear_distgroup are
# all disabled

my @disabled_noworld = (
    q{efs set_noworldaccess metaproj testmeta 1},
    q{efs set_noworldaccess project testmeta testproj 1},
    q{efs set_noworldaccess release testmeta testproj testrel 1},
);

foreach my $command (@disabled_noworld) {
    $test->efs_test_command(
        command => $command,
        not     => 1,
        expect  => [
            q{Unable to set the noworldaccess attribute, since the},
            q{disable_restricted_access feature has been enabled.},
        ],
    );
}

my @disabled_distgroup = (
    q{efs set_distgroup metaproj testmeta bogusgroup},
    q{efs set_distgroup project testmeta testproj bogusgroup},
    q{efs clear_distgroup metaproj testmeta},
    q{efs clear_distgroup project testmeta testproj},
);

foreach my $command (@disabled_distgroup) {
    $test->efs_test_command(
        command => $command,
        not     => 1,
        expect  => [
            q{Unable to set the distgroup attribute, since the},
            q{disable_restricted_access feature has been enabled.},
        ],
    );
}

# Verify that permissions get reverted as we expect

$test->efs_test_command(
    command => q{efs unlock release testmeta testproj testrel},
);

$test->efs_test_file(
    name  => q{Verify correct ownership/mode post-unlock},
    file  => qq{/efs/dev/$release/install/common/foo},
    user  => $username,
    group => $groupname,
    mode  => 0660,
);

$test->efs_test_command(
    command => qq{chmod o+r /efs/dev/$release/install/common/foo},
);

$test->efs_test_command(
    command => q{efs checkpoint release testmeta testproj testrel},
);

$test->efs_test_file(
    name  => q{Verify correct ownership/mode post-checkpoint},
    file  => qq{/efs/dev/$release/install/common/foo},
    user  => 0,
    group => 0,
    mode  => 0664,
);

# Verify that display commands don't display the noworldaccess and distgroup

my @display = (
    q{efs display metaproj testmeta},
    q{efs display project testmeta testproj},
);

foreach my $display (@display) {
    $test->efs_test_command(
        command => $display,
        filter  => qr{notify_list|noworldaccess|distgroup}msx,
        expect  => [ qr{notify_list}msx, ],
    );
}

# All done...

$test->efs_test_command(
    run_as  => q{efsops},
    command => q{efs undef feature disable_restricted_access},
    expect  => q{Undefined feature 'disable_restricted_access'},
);

$test->efs_test_command(
    command => q{efs history feature disable_restricted_access -count 2},
    expect  => [
        qr{History \s+ ID \s+ StartTime \s+ Duration \s+ Status \s+ Authuser \s+ Request}xms,
        qr{-+}xms,
        qr{define \s+ feature \s+ disable_restricted_access}xms,
        qr{undef \s+ feature \s+ disable_restricted_access}xms,
    ],
);

my @teardown = (
    q{efs setstage release testmeta testproj testrel rejected},
    q{efs destroy release testmeta testproj testrel },
    q{efs purge release testmeta testproj testrel -now},
);

foreach my $teardown (@teardown) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;

