#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More;
use File::Basename qw( basename );

use lib q{test/lib};
use EFS::Test::Config;

my $config = EFS::Test::Config->new;

my $sshcmd = $config->get_value( q{Global}, q{sshcmd} );

my $fstype = $config->get_value( q{Filesystem}, q{type} );

if ( $fstype eq q{afs} ) {
    plan skip_all => q{No need to teardown fake NFS infrastructure for AFS tests};
}

my $ldc_name      = $config->get_value( q{LocalDevCell}, q{name} );
my $ldc_shortname = $config->get_value( q{LocalDevCell}, q{shortname} );

my $rdc_name      = $config->get_value( q{RemoteDevCell}, q{name} );
my $rdc_shortname = $config->get_value( q{RemoteDevCell}, q{shortname} );

my $lpc_name      = $config->get_value( q{LocalProdCell}, q{name} );
my $lpc_shortname = $config->get_value( q{LocalProdCell}, q{shortname} );

my $ldc_efsserver  = $config->get_value( q{LocalDevCell}, q{efsserver} );
my $ldc_fileserver = $config->get_value( q{LocalDevCell}, q{fileserver} );

my $rdc_efsserver  = $config->get_value( q{RemoteDevCell}, q{efsserver} );
my $rdc_fileserver = $config->get_value( q{RemoteDevCell}, q{fileserver} );

my $lpc_efsserver  = $config->get_value( q{LocalProdCell}, q{efsserver} );
my $lpc_fileserver = $config->get_value( q{LocalProdCell}, q{fileserver} );

if ( $ldc_efsserver ne $ldc_fileserver ||
         $rdc_efsserver ne $rdc_fileserver ||
             $lpc_fileserver !~ /^fake/msx ) {
    die qq{Unable to bootstrap the efsservers if they use NFS servers (yet)\n};
}

my @localdirs = (
    q{/.efs},
    q{/.efsglobal},
    qq{/.efsadmin/$ldc_name/$ldc_fileserver/var/tmp},
    qq{/.efsadmin/$ldc_name/$ldc_fileserver/var},
    qq{/.efsadmin/$ldc_name/$ldc_fileserver},
    qq{/.efsadmin/$ldc_name},
    qq{/.efsadmin/$lpc_name/$lpc_fileserver/var/tmp},
    qq{/.efsadmin/$lpc_name/$lpc_fileserver/var},
    qq{/.efsadmin/$lpc_name/$lpc_fileserver},
    qq{/.efsadmin/$lpc_name},
    q{/.efsadmin},
);

my $ldc_devshare  = $config->get_value( q{LocalDevCell}, q{devshare} );
my $ldc_homeshare = $config->get_value( q{LocalDevCell},  q{homeshare} );
my $ldc_distshare = $config->get_value( q{LocalDevCell}, q{distshare} );
my $lpc_distshare = $config->get_value( q{LocalProdCell}, q{distshare} );

my @localshares = (
    $ldc_devshare,
    $ldc_homeshare,
    $ldc_distshare,
    $lpc_distshare,
);

my @locallinks = (
    q{/efs/dev},
    q{/efs/home},
    q{/efs/dist},
    qq{/.efsadmin/$ldc_name/$ldc_fileserver/$ldc_devshare},
    qq{/.efsadmin/$ldc_name/$ldc_fileserver/$ldc_homeshare},
    qq{/.efsadmin/$ldc_name/$ldc_fileserver/$ldc_distshare},
    qq{/.efsadmin/$lpc_name/$lpc_fileserver/$lpc_distshare},
    q{/.efsglobal/} . basename( $ldc_devshare ),
    q{/.efsglobal/} . basename( $ldc_homeshare ),
);

my @remotedirs = (
    q{/.efs},
    q{/.efsglobal},
    qq{/.efsadmin/$rdc_name/$rdc_fileserver/var/tmp},
    qq{/.efsadmin/$rdc_name/$rdc_fileserver/var},
    qq{/.efsadmin/$rdc_name/$rdc_fileserver},
    qq{/.efsadmin/$rdc_name},
    q{/.efsadmin},
);

if ( not -l q{/efs/test} ) {
    # NOTE: This assumes both hosts use the same style setup
    unshift @localdirs, q{/efs};
    unshift @remotedirs, q{/efs};
}

my $rdc_devshare  = $config->get_value( q{RemoteDevCell}, q{devshare} );
my $rdc_homeshare = $config->get_value( q{RemoteDevCell}, q{homeshare} );
my $rdc_distshare = $config->get_value( q{RemoteDevCell}, q{distshare} );

my @remoteshares = (
    $rdc_devshare,
    $rdc_homeshare,
    $rdc_distshare,
);

my @remotelinks = (
    q{/efs/dev},
    q{/efs/home},
    q{/efs/dist},
    qq{/.efsadmin/$rdc_name/$rdc_fileserver/$rdc_devshare},
    qq{/.efsadmin/$rdc_name/$rdc_fileserver/$rdc_homeshare},
    qq{/.efsadmin/$rdc_name/$rdc_fileserver/$rdc_distshare},
    q{/.efsglobal/} . basename( $rdc_devshare ),
    q{/.efsglobal/} . basename( $rdc_homeshare ),
);

foreach my $remotelink ( @remotelinks ) {
    system( qq{$sshcmd $rdc_efsserver rm -f $remotelink} );
    $CHILD_ERROR and die qq{Unable to unlink symlink $rdc_efsserver:$remotelink\n};
    ok( 1, qq{Removed link $rdc_efsserver:$remotelink} );
}

foreach my $remoteshare ( @remoteshares ) {
    system( qq{$sshcmd $rdc_efsserver rm -rf $remoteshare} );

    $CHILD_ERROR and die qq{Unable to rm -rf $rdc_efsserver:$remoteshare\n};
    ok( 1, qq{Removed share $rdc_efsserver:$remoteshare} );
}

foreach my $remotedir ( @remotedirs ) {
    system( qq{$sshcmd $rdc_efsserver rm -rf $remotedir} );
    $CHILD_ERROR and die qq{Unable to rm -rf $rdc_efsserver:$remotedir\n};
    ok( 1, qq{Removed directory $rdc_efsserver:$remotedir} );
}

foreach my $locallink ( @locallinks ) {
    if ( -l $locallink ) {
        unlink($locallink) ||
            die qq{Unable to unlink $locallink: $ERRNO\n};
        ok( 1, qq{unlink $locallink} );
    } else {
        ok( 1, qq{$locallink already gone} );
    }
}

foreach my $localshare ( @localshares ) {
    if ( -d $localshare ) {
        system(qq{rm -r $localshare});
        $CHILD_ERROR and die qq{Unable to rm -rf $localshare\n};
        ok( 1, qq{rm -r $localshare} );
    } else {
        ok( 1, qq{$localshare already gone} );
    }
}

foreach my $localdir ( @localdirs ) {
    if ( -d $localdir ) {
        rmdir($localdir) ||
            die qq{Unable to rmdir $localdir: $ERRNO\n};
        ok( 1, qq{rmdir $localdir} );
    } else {
        ok( 1, qq{$localdir already gone} );
    }
}

done_testing();
exit 0;

1;

__END__

=for :stopwords unmount teardown efsservers NAS symlink symlinks filesytems autofs efs

=head1 NOTES

This file will teardown the two efsservers, and return them to the
same state they were in before the test suite ran.  This will also
behave differently, depending on whether or not the dev/dist shares
are local directories, or real NAS shares.

=head2 Local Directories

=head3 Remove the symlink trees

All of the /efs, /.efs, /.efsglobal and /.efsadmin symlinks tree will
be removed completely.  A utility function that recursively wipes out
a directory, but ONLY if it contains nothing more than other
directories and symlinks, would be useful here.

=head3 Purge the contents of the dev/dist shares

Remove these directories entirely.  We own them, since we created them
to start with.

=head2 NAS Shares

=head3 Purge the contents of the NAS share

We need to be extra careful here, in case these volumes end up being
shared after all.  Instead of using C<rm -rf *> (the programmatic
equivalent), we should remove the specific directories we expect to be
there, and then raise an exception if we see anything unexpected.

=head3 Unmount /efs and /.efs filesytems

Pretty simple.

=head3 Remove the autofs configuration

Simply remove the /etc/auto.master entry and restart autofs.

=head2 Purge the local efs configuration

Simple remove /var/efs and /var/spool/efslogs.  We're going to
auto-generate the content every time in 00_setup/*.t

=cut
