package EFS::Test;

use strict;
use warnings;
use English;

use POSIX qw( setuid setgid );
use Text::Diff;
use Params::Validate qw( :all );
use Env::Path;
use Test::More;
use IO::Pipe;
use Readonly;
use Test::More;
use Carp;

# use AFS::Utils qw( setpag );

use EFS::Test::Config;

Readonly my $SPACE => q{ };

## no critic: RequireLocalizedPunctuationVars

BEGIN {

    Env::Path->PATH->Prepend(q{./bin});
    Env::Path->PERL5LIB->Prepend(q{./blib/lib});
    $ENV{ANSI_COLORS_DISABLED}  = 1;
    $ENV{EFS_TEST_SUITE}        = 1;
    $ENV{EFS_CLIENT_NO_WAITDIR} = 1;

    if ( -l q{/efs/test} ) {
        if ( not $ENV{EFS_PLATFORM} or not $ENV{EFS_PERL_DEPENDS_RUNTIME_GLOBAL} ) {
            die(
                qq{Both EFS_PLATFORM and EFS_PERL_DEPENDS_RUNTIME_GLOBAL\n},
                qq{environment variables must be set before running the\n},
                qq{EFS test suite using the /efs/test export:\n},
            );
        }
    }

}

sub new {

    my $proto = shift;
    my $class = ref $proto || $proto;

    my (%args) = validate( @_, {
        as_user => { type => SCALAR, optional => 1 },
        config  => { type => SCALAR, default => EFS::Test::Config->new, optional => 1 },
        verbose => { type => SCALAR, optional => 1, default => 0 },
    });

    my $self = bless {
        config  => $args{config},
        verbose => $args{verbose},
    }, $class;

    my $primary_user = $self->{config}->get_value( q{Global}, q{primary_user} );

    $self->{primary_user} = $self->get_user_data( $primary_user );
    $self->{efsops}       = $self->get_user_data( q{efsops} );

    if ( $args{as_user} ) {
        $self->become_user( $args{as_user} );
    }

    return $self;

}

sub get_user_data {

    my $self = shift;
    my $user = shift;

    my @getpwnam = getpwnam ($user) or
        die qq{No such username '$user'\n};

    my @getgrgid = getgrgid ( $getpwnam[3] ) or
        die qq{No such group ID '$getpwnam[3]'\n};

    my $data = {
        username  => $user,
        groupname => $getgrgid[0],
        uid       => $getpwnam[2],
        gid       => $getpwnam[3],
        home      => $getpwnam[7],
    };

    return $data;

}

sub get_primary_username {
    return shift->{primary_user}->{username};
}

sub get_primary_groupname {
    return shift->{primary_user}->{groupname};
}

sub get_primary_home {
    return shift->{primary_user}->{home};
}

sub get_config {
    return shift->{config};
}

sub passed {
    my $self = shift;
    my $name = shift;
    if ( not $ENV{EFS_TEST_NO_TAP} ) {
        ok( 1, $name );
    }
    return;
}

sub failed {
    my $self = shift;
    my $name = shift;
    if ( not $ENV{EFS_TEST_NO_TAP} ) {
        ok( 0, $name );
    }
    if ( $self->{verbose} ) {
        eval { confess 'stacktrace' };
        diag( $@ );
    }
    return;
}

sub efs_test_command {

    my $self = shift;

    my (%args) = validate( @_, {
        command   => { type     => SCALAR },
        name      => { type     => SCALAR,
                       optional => 1 },
        run_as    => { type     => SCALAR,
                       default  => q{primary_user},
                       optional => 1 },
        bail      => { type     => SCALAR,
                       optional => 1 },
        output    => { type     => ARRAYREF,
                       optional => 1 },
        expect    => { type     => SCALAR | SCALARREF | ARRAYREF,
                       optional => 1 },
        filter    => { isa      => q{Regexp},
                       depends  => [qw( expect )],
                       optional => 1 },
        filternot => { isa      => q{Regexp},
                       depends  => [qw( expect )], optional => 1 },
        not       => { type     => SCALAR,
                       default  => 0,
                       optional => 1 },
        verbose   => { type     => SCALAR,
                       default  => $PERLDB,
                       optional => 1 },
    });

    if ( $args{filter} && ref $args{filter} ne q{Regexp} ) {
        die qq{Not a Regexp: $args{filter}\n};
    }

    if ( not $args{name} ) {
        $args{name} = ( $args{not} ? q{!} : q{} ) . $args{command};
    }

    if ( $ENV{EFS_TEST_DIAG_OUTPUT} or $self->{verbose} ) {
        $args{verbose} = 1;
    }

    if ( $args{verbose} ) {
        diag( qq{\n\nefs_test_command args:} );
        diag( qq{     name: $args{name}} );
        diag( qq{  command: $args{command}} );
    }

    my $results = $self->command_output(
        command => $args{command},
        run_as  => $args{run_as},
    );

    my @output = @{ $results->{output} };
    my $status = $results->{status};

    if ( $args{output} ) {
        push @{ $args{output} }, @output;
    }

    if ( ( $status && $args{not} ) || ( $status == 0 && $args{not} == 0 ) ) {

        if ($status) {
            print qq{# Expected failure, got success\n};
        }
        else {
            print qq{# Expected success, got failure\n};
        }

        print qq{# Command output for diagnostics:\n};
        print map {qq{# $_\n}} @output;

        $self->failed( $args{name} );

        if ( $args{bail} ) {
            die qq{Bail out!\n};
        }
        else {
            return;
        }

    }

    my @output_complete = @output;

    if ( not $args{expect} ) {

        if ( $args{verbose} ) {
            diag( map { qq{$_\n} } @output_complete );
        }

        $self->passed( $args{name} );
        return 1;

    }

    if ( $args{filter} ) {
        @output = grep { $_ =~ $args{filter} } @output;
    }

    if ( $args{filternot} ) {
        @output = grep { $_ !~ $args{filternot} } @output;
    }

    my @expect
        = ref $args{expect} eq q{ARRAY}
        ? @{ $args{expect} }
        : ( $args{expect} );

    my $expected = q{};
    my $actual   = q{};

    my $diffs = 0;
    my $index = 0;

    while ( $index <= $#expect ) {

        my $expect = $expect[$index];

        if ( $index > $#output ) {
            $expected .= qq{$expect\n};
            $diffs++;
        }
        else {

            my $output         = $output[$index];
            my $output_compare = $output;

            # NOTE: Strip leading/trailing whitespace, and reduce sequences of
            # spaces to a single space, to simplify specifying matching output
            # in the tests
            $output_compare =~ s{^\s+}{}msx;    # Leading
            $output_compare =~ s{\s+$}{}msx;    # Trailing
            $output_compare =~ s{\s+}{$SPACE}gmsx; # 2+ to 1

            if ( ref $expect eq q{Regexp} ) {
                $actual .= qq{$output\n};
                if ( $output_compare !~ $expect ) {
                    $expected .= qq{$expect\n};
                    $diffs++;
                }
                else {
                    $expected .= qq{$output\n};
                }
            }
            else {

                my $expect_compare = $expect;
                $expect_compare =~ s{^\s+}{}msx;
                $expect_compare =~ s{\s+$}{}msx;
                $expect_compare =~ s{\s+}{$SPACE}gmsx;

                $expected .= qq{$expect_compare\n};
                $actual   .= qq{$output_compare\n};

                if ( $output_compare ne $expect_compare ) {
                    $diffs++;
                }

            }

        }

        $index++;

    }

    while ( $index <= $#output ) {
        my $output = $output[$index];
        $actual .= qq{$output\n};
        $diffs++;
        $index++;
    }

    if ( $diffs ) {

        my $text = diff \$actual, \$expected,
            {
                CONTEXT     => 3,
                STYLE       => q{Table},
                FILENAME_A  => q{Got},
                FILENAME_B  => q{Expected},
                OFFSET_A    => 1,
                OFFSET_B    => 1,
                INDEX_LABEL => q{Ln},
            };

        chomp ($text);
        $text .= qq{\n};
        $text =~ s/^/# /gmsx;
        print $text;

        $self->failed( $args{name} );

        if ( $args{bail} ) {
            die qq{Bail out!\n};
        }
        else {
            return;
        }

    }
    else {

        if ( $args{verbose} ) {
            diag( map { qq{$_\n} } @output_complete );
        }

        $self->passed( $args{name} );
        return 1;

    }

}

sub efs_test_link {

    my $self = shift;

    my (%args) = validate( @_, {
        link   => { type => SCALAR },
        target => { type => SCALAR },
        name   => { type => SCALAR, optional => 1 },
        bail   => { type => SCALAR, optional => 1 },
        not    => { type => SCALAR, default => 0, optional => 1 },
        user   => { type => SCALAR, default => 0, optional => 1 },
        group  => { type => SCALAR, default => 0, optional => 1 },
    });

    $args{path} = delete $args{link};
    $args{type} = q{link};

    return $self->efs_test_path(%args);

}

sub efs_test_directory {

    my $self = shift;

    my (%args) = validate( @_, {
        directory => { type => SCALAR },
        name      => { type => SCALAR, optional => 1 },
        bail      => { type => SCALAR, optional => 1 },
        not       => { type => SCALAR, default => 0, optional => 1 },
        user      => { type => SCALAR, default => 0, optional => 1 },
        group     => { type => SCALAR, default => 0, optional => 1 },
        mode      => { type => SCALAR, default => 0, optional => 1 },
        mask      => { type => SCALAR, default => 0777, optional => 1 },
        nolink    => { type => SCALAR, optional => 1 },
    });

    $args{path} = delete $args{directory};
    $args{type} = q{dir};

    return $self->efs_test_path(%args);

}

sub efs_test_file {

    my $self = shift;

    my (%args) = validate( @_, {
        file  => { type => SCALAR },
        name  => { type => SCALAR, optional => 1 },
        bail  => { type => SCALAR, default => 0, optional => 1 },
        not   => { type => SCALAR, default => 0, optional => 1 },
        user  => { type => SCALAR, default => 0, optional => 1 },
        group => { type => SCALAR, default => 0, optional => 1 },
        mode  => { type => SCALAR, default => 0, optional => 1 },
        mask  => { type => SCALAR, default => 0777, optional => 1 },
    });

    $args{path} = delete $args{file};
    $args{type} = q{file};

    return $self->efs_test_path(%args);

}

sub efs_test_path {

    my $self = shift;

    my (%args) = @_;

    if ( not $args{name} ) {
        $args{name} = ( $args{not} ? q{!} : q{} ) . qq{$args{type} $args{path}};
        $args{name} .= qq{ -> $args{target}} if $args{type} eq q{link};
    }

    my $exists = 0;

    if ( $args{type} eq q{file} ) {
        $exists = -f $args{path};
    }
    elsif ( $args{type} eq q{dir} ) {
        $exists = -d $args{path};
        if ( $args{nolink} and -l $args{path} ) {
            STDOUT->print(
                qq{# Path '$args{path}' is a link, not a directory\n},
            );
            $exists = 0;
        }
    }
    elsif ( $args{type} eq q{link} ) {
        $exists = -l $args{path};
    }
    else {
        die qq{Unsupported path type '$args{type}'\n};
    }

    if ( not $exists ) {
        if ( $args{not} ) {
            $self->passed( $args{name} );
            return 1;
        }
        else {
            STDOUT->print(
                qq{# Path '$args{path}' doesn't appear to exist\n},
            );
            $self->failed( $args{name} );
            if ( $args{bail} ) {
                die qq{Bail out!\n};
            }
            else {
                return;
            }
        }
    }

    my $correct_ownership = 0;
    my $correct_mode      = 0;

    if ( $args{type} eq q{link} ) {

        my $readlink = readlink ( $args{path} )
            || die qq{Unable to readlink $args{path}: $ERRNO\n};

        if ( $readlink eq $args{target} ) {
            $correct_ownership = 1;
            $correct_mode      = 1;
        }
        else {
            STDOUT->print(
                qq{# Link '$args{path}' has target '$readlink', } .
                    qq{instead of '$args{target}'\n},
            );
        }

    }
    else {

        $correct_ownership = $self->verify_ownership(
            path  => $args{path},
            user  => $args{user},
            group => $args{group},
        );

        if ( $args{mode} ) {
            $correct_mode = $self->verify_mode(
                path => $args{path},
                mode => $args{mode},
                mask => $args{mask},
            );
        }
        else {
            $correct_mode = 1;
        }

    }

    if ( $correct_ownership && $correct_mode ) {
        $self->passed( $args{name} );
        return 1;
    }
    else {
        $self->failed( $args{name} );
        if ( $args{bail} ) {
            die qq{Bail out!\n};
        }
        else {
            return;
        }
    }

}

sub verify_ownership {

    my $self = shift;

    my (%args) = validate( @_, {
        path  => { type => SCALAR },
        user  => { type => SCALAR },
        group => { type => SCALAR },
    });

    my ( $uid, $gid ) = ( 0, 0 );

    if ( $args{user} =~ /^\d+$/msx ) {
        $uid = $args{user};
    }

    else {
        my @user = getpwnam ( $args{user} )
            or die qq{No such user '$args{user}'\n};
        $uid = $user[2];
    }

    if ( $args{group} =~ /^\d+$/msx ) {
        $gid = $args{group};
    }
    elsif ( $args{group} eq q{root} ) {
        $gid = 0;
    }
    else {
        my @group = getgrnam ( $args{group} )
            or die qq{No such group '$args{group}'\n};
        $gid = $group[2];
    }

    my @stat = stat ( $args{path} )
        or die qq{Unable to stat $args{path}: $ERRNO\n};

    if ( $stat[4] == $uid && $stat[5] == $gid ) {
        return 1;
    }
    else {
        STDOUT->print(
            qq{# Path $args{path} has uid/gid mismatch:\n},
            qq{# Should be uid == $uid, gid == $gid\n},
            qq{# But has uid == $stat[4], gid == $stat[5]\n},
        );
        return;
    }

}

sub verify_mode {

    my $self = shift;

    my (%args) = validate( @_, {
        path => { type => SCALAR },
        mode => { type => SCALAR },
        mask => { type => SCALAR },
    });

    my @stat = stat ( $args{path} )
        or die qq{Unable to stat $args{path}: $ERRNO\n};

    my $mode = $stat[2] & $args{mask};

    if ( $mode ^ $args{mode} ) {
        STDOUT->printf(
            qq{# Path $args{path} has mode of %o, but should be %o\n},
            $mode, $args{mode}, );
        return;
    }
    else {
        return 1;
    }

}

sub become_user {

    my $self = shift;
    my $user = shift;

    if ( $EUID != 0 ) {
        die qq{Unable to run as primary user: not running as root\n};
    }

    if ( not $user ~~ [qw( primary_user efsops )] ) {
        die qq{Unable to run as other user: must be 'efsops' or 'primary_user'\n};
    }

    setgid( $self->{$user}->{gid} ) ||
        die qq{Unable to setgid: $ERRNO\n};

    setuid( $self->{$user}->{uid} ) ||
        die qq{Unable to setuid: $ERRNO\n};

    $ENV{HOME} = $self->{$user}->{home};

    my $username = $self->{$user}->{username};

    my $config = $self->get_config;

    if ( $config->get_value( q{Security}, q{kerberos_realm} ) ) {

        $ENV{KRB5CCNAME} = qq{FILE:/tmp/krb5cc_efs_${username}_$PID};
        system qq{kinit -k -t /var/kerberos/keytabs/$username.keytab $username};
        $CHILD_ERROR and die qq{Unable to acquire kerberos credentials for $username\n};

        # XXX: Needs to be configurable
        # setpag or die qq{Unable to acquire new AFS PAG: $ERRNO\n};

        # XXX: List of cells?  efsd needs that, since it manages numerous
        # cells, but the efs client should just be modifying paths in the
        # localcell.  We'll see...
        system q{aklog};
        $CHILD_ERROR and die qq{Unable to acquire AFS tokens using aklog\n};
        
    }

    $self->{as_user} = $user;

    return;

}

sub run_as_efsops {

    my $self = shift;
    my $name = shift;
    my $code = shift;

    return $self->run_as( $name, $code, q{efsops} );

}

sub run_as_primary_user {

    my $self = shift;
    my $name = shift;
    my $code = shift;

    return $self->run_as( $name, $code, q{primary_user} );

}

sub run_as {

    my $self = shift;
    my $name = shift;
    my $code = shift;
    my $user = shift;

    if ( $self->{as_user} ) {
        die qq{Object was instantiated with 'as_user'\n};
    }

    my $pid = $PERLDB ? 0 : fork;

    if ( not defined $pid ) {
        die qq{Unable to fork: $ERRNO\n};
    }

    my $status = 0;

    if ( $pid == 0 ) {
        local $ENV{EFS_TEST_NO_TAP} = 1;
        $self->become_user( $user );
        exit( $code->() ? 0 : 1 );
    } else {
        if ( waitpid( $pid, 0 ) == -1 ) {
            die qq{Error reaping $pid: $ERRNO\n};
        }
        $status = ( $CHILD_ERROR == 0 ? 1 : 0 );
        ok( $status, $name );
    }

    return $status;

}

sub command_output {

    my $self = shift;
    my (%args) = @_;

    my $command = $args{command};

    if ( $self->{as_user} || $args{run_as} eq q{root} ) {

        my @output = qx{ $args{command} 2>&1 };
        chomp @output;

        @output = grep { not m{^\s*$}msx } @output;

        my $status = not $CHILD_ERROR;

        return {
            status => $status,
            output => \@output,
        };

    }

    my $results = {
        output => [],
        status => 0,
    };

    my $pipe = IO::Pipe->new || die qq{Unable to create pipe: $ERRNO\n};
    my $pid = fork;

    if ( not defined $pid ) {
        die qq{Unable to fork: $ERRNO\n};
    }

    if ( $pid == 0 ) {

        $self->become_user( $args{run_as} );

        STDOUT->fdopen( $pipe->writer->fileno, q{w} )
            or die qq{Error reopening STDOUT\n};
        STDERR->fdopen( STDOUT->fileno, q{w} )
            or die qq{Error reopening STDOUT\n};
        STDOUT->autoflush(1);
        STDERR->autoflush(1);

        exec $args{command};

    }
    else {

        $pipe->reader;

        while ( defined( $_ = $pipe->getline) ) {
            chomp;
            next if m{^\s*$}msx;
            push @{ $results->{output} }, $_;
        }

    }

    if ( not waitpid( $pid, 0 ) ) {
        warn qq{Unable to waitpid child process ($pid)\n};
        $results->{status} = undef;
    }

    $results->{status} = ( $CHILD_ERROR ? 0 : 1 );

    return $results;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
